import sqlalchemy as sa
from pytest import mark, raises
from sqlalchemy_utils import (
    ChoiceType,
    ColorType,
    CountryType,
    EmailType,
    IntRangeType,
    PasswordType,
    PhoneNumberType,
    URLType,
    UUIDType
)
from sqlalchemy_utils.types import arrow, phone_number, WeekDaysType  # noqa
from wtforms.fields import (
    BooleanField,
    FloatField,
    PasswordField,
    TextAreaField
)
from wtforms.validators import Length, URL
from wtforms_components import Email
from wtforms_components.fields import (
    ColorField,
    DateField,
    DateTimeField,
    DecimalField,
    EmailField,
    IntegerField,
    IntIntervalField,
    SelectField,
    StringField,
    TimeField
)

from tests import ModelFormTestCase
from wtforms_alchemy import (
    CountryField,
    ModelForm,
    null_or_unicode,
    PhoneNumberField,
    UnknownTypeException,
    WeekDaysField
)
from wtforms_alchemy.utils import ClassMap

passlib = None
try:
    import passlib  # noqa
except ImportError:
    pass


class UnknownType(sa.types.UserDefinedType):

    def get_col_spec(self):
        return "UNKNOWN()"


class CustomUnicodeTextType(sa.types.TypeDecorator):
    impl = sa.types.UnicodeText


class CustomUnicodeType(sa.types.TypeDecorator):
    impl = sa.types.Unicode


class CustomNumericType(sa.types.TypeDecorator):
    impl = sa.types.Numeric


class TestModelColumnToFormFieldTypeConversion(ModelFormTestCase):
    def test_raises_exception_for_unknown_type(self):
        with raises(UnknownTypeException):
            self.init(type_=UnknownType)
            self.form_class()

    def test_unicode_converts_to_text_field(self):
        self.init()
        self.assert_type('test_column', StringField)

    def test_custom_unicode_converts_to_text_field(self):
        self.init(type_=CustomUnicodeType)
        self.assert_type('test_column', StringField)

    def test_string_converts_to_text_field(self):
        self.init(type_=sa.String)
        self.assert_type('test_column', StringField)

    def test_integer_converts_to_integer_field(self):
        self.init(type_=sa.Integer)
        self.assert_type('test_column', IntegerField)

    def test_unicode_text_converts_to_text_area_field(self):
        self.init(type_=sa.UnicodeText)
        self.assert_type('test_column', TextAreaField)

    def test_custom_unicode_text_converts_to_text_area_field(self):
        self.init(type_=CustomUnicodeTextType)
        self.assert_type('test_column', TextAreaField)

    def test_boolean_converts_to_boolean_field(self):
        self.init(type_=sa.Boolean)
        self.assert_type('test_column', BooleanField)

    def test_datetime_converts_to_datetime_field(self):
        self.init(type_=sa.DateTime)
        self.assert_type('test_column', DateTimeField)

    def test_date_converts_to_date_field(self):
        self.init(type_=sa.Date)
        self.assert_type('test_column', DateField)

    def test_float_converts_to_float_field(self):
        self.init(type_=sa.Float)
        self.assert_type('test_column', FloatField)

    def test_numeric_converts_to_decimal_field(self):
        self.init(type_=sa.Numeric)
        self.assert_type('test_column', DecimalField)

    def test_numeric_scale_converts_to_decimal_field_scale(self):
        self.init(type_=sa.Numeric(scale=4))
        form = self.form_class()
        assert form.test_column.places == 4

    def test_custom_numeric_converts_to_decimal_field(self):
        self.init(type_=CustomNumericType)
        self.assert_type('test_column', DecimalField)

    def test_enum_field_converts_to_select_field(self):
        choices = ['1', '2']
        self.init(type_=sa.Enum(*choices))
        self.assert_type('test_column', SelectField)
        form = self.form_class()
        assert form.test_column.choices == [(s, s) for s in choices]

    def test_nullable_enum_uses_null_or_unicode_coerce_func_by_default(self):
        choices = ['1', '2']
        self.init(type_=sa.Enum(*choices), nullable=True)
        field = self._get_field('test_column')
        assert field.coerce == null_or_unicode

    def test_custom_choices_override_enum_choices(self):
        choices = ['1', '2']
        custom_choices = [('2', '2'), ('3', '3')]
        self.init(type_=sa.Enum(*choices), info={'choices': custom_choices})
        form = self.form_class()
        assert form.test_column.choices == custom_choices

    def test_column_with_choices_converts_to_select_field(self):
        choices = [(u'1', '1'), (u'2', '2')]
        self.init(type_=sa.Integer, info={'choices': choices})
        self.assert_type('test_column', SelectField)
        form = self.form_class()
        assert form.test_column.choices == choices

    def test_assigns_email_validator_for_email_type(self):
        self.init(type_=EmailType)
        self.assert_has_validator('test_column', Email)

    def test_assigns_url_validator_for_url_type(self):
        self.init(type_=URLType)
        self.assert_has_validator('test_column', URL)

    def test_time_converts_to_time_field(self):
        self.init(type_=sa.types.Time)
        self.assert_type('test_column', TimeField)

    def test_varchar_converts_to_text_field(self):
        self.init(type_=sa.types.VARCHAR)
        self.assert_type('test_column', StringField)

    def test_text_converts_to_textarea_field(self):
        self.init(type_=sa.types.TEXT)
        self.assert_type('test_column', TextAreaField)

    def test_char_converts_to_text_field(self):
        self.init(type_=sa.types.CHAR)
        self.assert_type('test_column', StringField)

    def test_real_converts_to_float_field(self):
        self.init(type_=sa.types.REAL)
        self.assert_type('test_column', FloatField)

    @mark.xfail('phone_number.phonenumbers is None')
    def test_phone_number_converts_to_phone_number_field(self):
        self.init(type_=PhoneNumberType)
        self.assert_type('test_column', PhoneNumberField)

    @mark.xfail('phone_number.phonenumbers is None')
    def test_phone_number_country_code_passed_to_field(self):
        self.init(type_=PhoneNumberType(region='SE'))
        form = self.form_class()
        assert form.test_column.region == 'SE'

    @mark.xfail('phone_number.phonenumbers is None')
    def test_phone_number_type_has_no_length_validation(self):
        self.init(type_=PhoneNumberType(country_code='FI'))
        field = self._get_field('test_column')
        for validator in field.validators:
            assert validator.__class__ != Length

    @mark.parametrize(('type', 'field'), (
        (IntRangeType, IntIntervalField),
    ))
    def test_range_type_conversion(self, type, field):
        self.init(type_=type)
        self.assert_type('test_column', field)

    @mark.xfail('passlib is None')
    def test_password_type_converts_to_password_field(self):
        self.init(type_=PasswordType)
        self.assert_type('test_column', PasswordField)

    @mark.xfail('arrow.arrow is None')
    def test_arrow_type_converts_to_datetime_field(self):
        self.init(type_=arrow.ArrowType)
        self.assert_type('test_column', DateTimeField)

    def test_url_type_converts_to_string_field(self):
        self.init(type_=URLType)
        self.assert_type('test_column', StringField)

    def test_uuid_type_converst_to_uuid_type(self):
        self.init(type_=UUIDType)
        self.assert_type('test_column', StringField)

    def test_color_type_converts_to_color_field(self):
        self.init(type_=ColorType)
        self.assert_type('test_column', ColorField)

    def test_email_type_converts_to_email_field(self):
        self.init(type_=EmailType)
        self.assert_type('test_column', EmailField)

    def test_country_type_converts_to_country_field(self):
        self.init(type_=CountryType)
        self.assert_type('test_column', CountryField)

    def test_choice_type_converts_to_select_field(self):
        choices = [(u'1', u'choice 1'), (u'2', u'choice 2')]
        self.init(type_=ChoiceType(choices))
        self.assert_type('test_column', SelectField)
        assert self.form_class().test_column.choices == choices

    def test_choice_type_uses_custom_coerce_func(self):
        choices = [(u'1', u'choice 1'), (u'2', u'choice 2')]
        self.init(type_=ChoiceType(choices))
        self.assert_type('test_column', SelectField)
        model = self.ModelTest(test_column=u'2')
        form = self.form_class(obj=model)
        assert '<option selected value="2">' in str(form.test_column)


class TestWeekDaysTypeConversion(ModelFormTestCase):

    def test_weekdays_type_converts_to_weekdays_field(self):
        self.init(type_=WeekDaysType)
        self.assert_type('test_column', WeekDaysField)


class TestCustomTypeMap(ModelFormTestCase):
    def test_override_type_map_on_class_level(self):
        class ModelTest(self.base):
            __tablename__ = 'model_test'
            id = sa.Column(sa.Integer, primary_key=True)
            test_column = sa.Column(sa.Unicode(255), nullable=False)

        class ModelTestForm(ModelForm):
            class Meta:
                model = ModelTest
                not_null_validator = None
                type_map = ClassMap({sa.Unicode: TextAreaField})

        form = ModelTestForm()
        assert isinstance(form.test_column, TextAreaField)

    def test_override_type_map_with_callable(self):
        class ModelTest(self.base):
            __tablename__ = 'model_test'
            id = sa.Column(sa.Integer, primary_key=True)
            test_column_short = sa.Column(sa.Unicode(255), nullable=False)
            test_column_long = sa.Column(sa.Unicode(), nullable=False)

        class ModelTestForm(ModelForm):
            class Meta:
                model = ModelTest
                not_null_validator = None
                type_map = ClassMap({
                    sa.Unicode: lambda column: (
                        StringField if column.type.length else TextAreaField
                    )
                })

        form = ModelTestForm()
        assert isinstance(form.test_column_short, StringField)
        assert isinstance(form.test_column_long, TextAreaField)



#!/usr/bin/env python
# encoding: utf-8
#
# rss.py
#
# Licensed under a 3-clause BSD license.
#
# Revision history:
#     11 Apr 2016 J. Sánchez-Gallego
#       Initial version


from __future__ import division
from __future__ import print_function

import warnings

from astropy.io import fits
import numpy as np

import marvin

from marvin.core.core import MarvinToolsClass
from marvin.core.exceptions import MarvinError, MarvinUserWarning

from marvin.tools.spectrum import Spectrum


class RSS(MarvinToolsClass, list):
    """A class to interface with MaNGA RSS data.

    This class represents a fully reduced RSS file, initialised either
    from a file, a database, or remotely via the Marvin API. The class
    inherits from Python's list class, and is defined as a list of
    RSSFiber objects.

    Parameters:
        filename (str):
            The path of the file containing the RSS to load.
        mangaid (str):
            The mangaid of the RSS to load.
        plateifu (str):
            The plate-ifu of the RSS to load (either ``mangaid`` or
            ``plateifu`` can be used, but not both).
        mode ({'local', 'remote', 'auto'}):
            The load mode to use. See
            :doc:`Mode secision tree</mode_decision>`.
        nsa_source ({'auto', 'drpall', 'nsa'}):
            Defines how the NSA data for this object should loaded when
            ``RSS.nsa`` is first called. If ``drpall``, the drpall file will
            be used (note that this will only contain a subset of all the NSA
            information); if ``nsa``, the full set of data from the DB will be
            retrieved. If the drpall file or a database are not available, a
            remote API call will be attempted. If ``nsa_source='auto'``, the
            source will depend on how the ``RSS`` object has been
            instantiated. If the cube has ``RSS.data_origin='file'``,
            the drpall file will be used (as it is more likely that the user
            has that file in their system). Otherwise, ``nsa_source='nsa'``
            will be assumed. This behaviour can be modified during runtime by
            modifying the ``RSS.nsa_mode`` with one of the valid values.
        release (str):
            The MPL/DR version of the data to use.

    Return:
        rss:
            An object representing the RSS entity. The object is a list of
            RSSFiber objects, one for each fibre in the RSS entity.

    """

    def __init__(self, *args, **kwargs):

        valid_kwargs = ['filename', 'mangaid', 'plateifu', 'mode',
                        'drpall', 'release', 'nsa_source']

        assert len(args) == 0, 'RSS does not accept arguments, only keywords.'
        for kw in kwargs:
            assert kw in valid_kwargs, 'keyword {0} is not valid'.format(kw)

        self.data = None

        MarvinToolsClass.__init__(self, *args, **kwargs)

        if self.data_origin == 'file':
            self._load_rss_from_file()
        elif self.data_origin == 'db':
            self._load_rss_from_db()
        elif self.data_origin == 'api':
            self._load_rss_from_api()

        _fibers = self._init_fibers()
        list.__init__(self, _fibers)

        # TODO: check that the drpver of the loaded data matches the one in the object.

    def __repr__(self):
        """Representation for RSS."""

        return ('<Marvin RSS (mangaid={self.mangaid!r}, plateifu={self.plateifu!r}, '
                'mode={self.mode!r}, data_origin={self.data_origin!r})>'.format(self=self))

    def _getFullPath(self):
        """Returns the full path of the file in the tree."""

        if not self.plateifu:
            return None

        plate, ifu = self.plateifu.split('-')

        return super(RSS, self)._getFullPath('mangarss', ifu=ifu, drpver=self._drpver, plate=plate)

    def download(self):
        """Downloads the cube using sdss_access - Rsync"""

        if not self.plateifu:
            return None

        plate, ifu = self.plateifu.split('-')

        return super(RSS, self).download('mangarss', ifu=ifu, drpver=self._drpver, plate=plate)

    def _load_rss_from_file(self):
        """Initialises the RSS object from a file."""

        try:
            self.data = fits.open(self.filename)
            self.mangaid = self.data[0].header['MANGAID'].strip()
            self.plateifu = '{0}-{1}'.format(
                self.data[0].header['PLATEID'], self.data[0].header['IFUDSGN'])
        except Exception as ee:
            raise MarvinError('Could not initialize via filename: {0}'.format(ee))

        # Checks and populates release.
        file_drpver = self.data[0].header['VERSDRP3']
        file_drpver = 'v1_5_1' if file_drpver == 'v1_5_0' else file_drpver

        file_ver = marvin.config.lookUpRelease(file_drpver)
        assert file_ver is not None, 'cannot find file version.'

        if file_ver != self._release:
            warnings.warn('mismatch between file version={0} and object release={1}. '
                          'Setting object release to {0}'.format(file_ver, self._release),
                          MarvinUserWarning)
            self._release = file_ver

        self._drpver, self._dapver = marvin.config.lookUpVersions(release=self._release)

    def _load_rss_from_db(self):
        """Initialises the RSS object from the DB."""

        import sqlalchemy

        mdb = marvin.marvindb

        if not mdb.isdbconnected:
            raise MarvinError('No db connected')

        plate, ifudesign = [item.strip() for item in self.plateifu.split('-')]

        try:
            self.data = mdb.session.query(mdb.datadb.RssFiber).join(
                mdb.datadb.Cube, mdb.datadb.PipelineInfo,
                mdb.datadb.PipelineVersion, mdb.datadb.IFUDesign).filter(
                    mdb.datadb.PipelineVersion.version == self._drpver,
                    mdb.datadb.Cube.plate == plate,
                    mdb.datadb.IFUDesign.name == ifudesign).all()

        except sqlalchemy.orm.exc.NoResultFound as ee:
            raise MarvinError('Could not retrieve RSS for plate-ifu {0}: '
                              'No Results Found: {1}'
                              .format(self.plateifu, ee))

        except Exception as ee:
            raise MarvinError('Could not retrieve RSS for plate-ifu {0}: '
                              'Unknown exception: {1}'
                              .format(self.plateifu, ee))

        if not self.data:
            raise MarvinError('Could not retrieve RSS for plate-ifu {0}: '
                              'Unknown error.'.format(self.plateifu))

    def _load_rss_from_api(self):
        """Initialises the RSS object using the remote API."""

        # Checks that the RSS exists.
        routeparams = {'name': self.plateifu}
        url = marvin.config.urlmap['api']['getRSS']['url'].format(**routeparams)

        # Make the API call
        self._toolInteraction(url)

    def _init_fibers(self):
        """Initialises the object as a list of RSSFiber instances."""

        if self.data_origin == 'file':
            _fibers = [RSSFiber._init_from_hdu(hdulist=self.data, index=ii)
                       for ii in range(self.data[1].data.shape[0])]

        elif self.data_origin == 'db':
            _fibers = [RSSFiber._init_from_db(rssfiber=rssfiber) for rssfiber in self.data]

        elif self.data_origin == 'api':
            # Makes a call to the API to retrieve all the arrays for all the fibres.

            routeparams = {'name': self.plateifu}
            url = marvin.config.urlmap['api']['getRSSAllFibers']['url'].format(**routeparams)

            # Make the API call
            response = self._toolInteraction(url)
            data = response.getData()

            wavelength = np.array(data['wavelength'])

            _fibers = []
            for ii in range(len(data) - 1):
                flux = np.array(data[str(ii)][0])
                ivar = np.array(data[str(ii)][1])
                mask = np.array(data[str(ii)][2])
                _fibers.append(
                    RSSFiber(flux, ivar=ivar, mask=mask, wavelength=wavelength,
                             mangaid=self.mangaid, plateifu=self.plateifu, data_origin='api'))

        return _fibers


class RSSFiber(Spectrum):
    """A class to represent a MaNGA RSS fiber.

    This class is basically a subclass of |spectrum| with additional
    functionality. It is not intended to be initialised directly, but via
    the :py:meth:`RSS._initFibers` method.

    Parameters:
        args:
            Arguments to pass to |spectrum| for initialisation.
        kwargs:
            Keyword arguments to pass to |spectrum| for initialisation.

    Return:
        rssfiber:
            An object representing the RSS fiber entity.

    .. |spectrum| replace:: :class:`~marvin.tools.spectrum.Spectrum`

    """

    def __init__(self, *args, **kwargs):

        self.mangaid = kwargs.pop('mangaid', None)
        self.plateifu = kwargs.pop('plateifu', None)
        self.data_origin = kwargs.pop('data_origin', None)

        flux_units = '1e-17 erg/s/cm^2/Ang/fiber'
        wavelength_unit = 'Angstrom'
        kwargs['units'] = flux_units
        kwargs['wavelength_unit'] = wavelength_unit

        Spectrum.__init__(self, *args, **kwargs)

    def __repr__(self):
        """Representation for RSSFiber."""

        return ('<Marvin RSSFiber (mangaid={self.mangaid!r}, plateifu={self.plateifu!r}, '
                'data_origin={self.data_origin!r})>'.format(self=self))

    @classmethod
    def _init_from_hdu(cls, hdulist, index):
        """Initialises a RSSFiber object from a RSS HDUList."""

        assert index is not None, \
            'if hdu is defined, an index is required.'

        mangaid = hdulist[0].header['MANGAID'].strip()
        plateifu = '{0}-{1}'.format(hdulist[0].header['PLATEID'], hdulist[0].header['IFUDSGN'])

        flux = hdulist['FLUX'].data[index, :]
        ivar = hdulist['IVAR'].data[index, :]
        mask = hdulist['MASK'].data[index, :]
        wave = hdulist['WAVE'].data

        obj = RSSFiber(flux, ivar=ivar, mask=mask, wavelength=wave, mangaid=mangaid,
                       plateifu=plateifu, data_origin='file')

        return obj

    @classmethod
    def _init_from_db(cls, rssfiber):
        """Initialites a RSS fiber from the DB."""

        mangaid = rssfiber.cube.mangaid
        plateifu = '{0}-{1}'.format(rssfiber.cube.plate, rssfiber.cube.ifu.name)

        obj = RSSFiber(rssfiber.flux, ivar=rssfiber.ivar, mask=rssfiber.mask,
                       wavelength=rssfiber.cube.wavelength.wavelength, mangaid=mangaid,
                       plateifu=plateifu, data_origin='db')

        return obj




import imp
import os
import sys
import warnings


def _import_module(name, relpath):
    """Imports a module from a path relative to this one's."""

    try:
        fp, pathname, description = imp.find_module(name)
        return imp.load_module(name, fp, pathname, description)
    except ImportError:
        pass

    try:
        path = os.path.join(os.path.dirname(__file__), relpath)
        fp, filename, description = imp.find_module(name, [path])
        return imp.load_module(name, fp, filename, description)
    except IOError:
        warnings.warn('Marvin cannot import {0}'.format(name), ImportWarning)
        return None


# Imports the external packages (some of them don't have __init__ in the right places and can
# not be imported normally).
# sdss_access = _import_module('sdss_access', 'sdss_access/python/')
sqlalchemy_boolean_search = _import_module(
    'sqlalchemy_boolean_search', 'sqlalchemy-boolean-search/')
wtforms_alchemy = _import_module('wtforms_alchemy', 'wtforms-alchemy/')
brain = _import_module('brain', 'marvin_brain/python')



from fabric.operations import local
from fabric.api import cd, env, task, prefix, run
from contextlib import contextmanager

@task
def runserver():
    local('python runserver.py')

@task
def test():
    # Requires "pip install pytest"
    local('py.test flask_user/tests/')

@task
def coverage():
    # Requires "pip install pytest-coverage"
    local('py.test --cov flask_user --cov-report term-missing --cov-config flask_user/tests/.coveragerc flask_user/tests/')

@task
def babel():
    local('pybabel extract -F flask_user/translations/babel.cfg -c NOTE -o flask_user/translations/flask_user.pot flask_user flask_user')
    for code in ('en', 'fi', 'fr', 'nl', 'sv', 'zh'):
        local('pybabel update -i flask_user/translations/flask_user.pot --domain=flask_user --output-dir flask_user/translations -l '+code)
    local('pybabel compile -f --domain=flask_user --directory flask_user/translations')

@task
def babel_init():
    local('pybabel extract -F flask_user/translations/babel.cfg -c NOTE -o flask_user/translations/flask_user.pot flask_user flask_user')
    for code in ('en', 'fi', 'fr', 'nl', 'sv', 'zh'):
        local('pybabel init -i flask_user/translations/flask_user.pot --domain=flask_user --output-dir flask_user/translations -l '+code)
    local('pybabel compile -f --domain=flask_user --directory flask_user/translations')

@task
def docs():
    local('cp example_apps/*_app.py docs/source/includes/.')
    local('sphinx-build -b html docs/source ../builds/flask_user/docs')
    local('cd ../builds/flask_user/docs && zip -u -r flask_user_docs *')

@task
def rebuild_docs():
    local('rm -fr ../builds/flask_user/docs')
    docs()

@task
def upload_to_pypi():
    local('python setup.py sdist upload')


#!/usr/bin/env python
# encoding: utf-8

import inspect
from functools import wraps

from decorator import decorator

import pytest


def use_bintypes(*bintypes):
    """Decorate test to run only for the given bintypes."""
    def check_bintype(f):
        @wraps(f)
        def decorated_function(self, *args, **kwargs):
            if kwargs['galaxy'].bintype not in bintypes:
                pytest.skip('Only use {}'.format(', '.join(bintypes)))
            return f(self, *args, **kwargs)
        return decorated_function
    return check_bintype


def use_releases(*releases):
    """Decorate test to run only for the given releases."""
    def check_bintype(f):
        @wraps(f)
        def decorated_function(self, *args, **kwargs):
            if 'release' in kwargs.keys():
                release = kwargs['release']
            elif 'galaxy' in kwargs.keys():
                release = kwargs['galaxy'].release
            if release not in releases:
                pytest.skip('Only use {}'.format(', '.join(releases)))
            return f(self, *args, **kwargs)
        return decorated_function
    return check_bintype


class MetaUse(object):
    """Meta class to define a testing class that decorates all tests to use the specified fxn."""
    def __init__(self, *args):
        self.args = args

    def __call__(self, decorated_class):
        for attr in inspect.getmembers(decorated_class, inspect.isfunction):
            # only decorate public functions
            if attr[0][0] != '_':
                setattr(decorated_class, attr[0],
                        self.fxn(*self.args)(attr[1]))
        return decorated_class


class UseBintypes(MetaUse):
    def __init__(self, *args):
        self.args = args
        self.fxn = use_bintypes


class UseReleases(MetaUse):
    def __init__(self, *args):
        self.args = args
        self.fxn = use_releases


# These decorators for functions and classes allow to skip or run tests only for galaxies
# that have certain bintypes, templates, or releases
def marvin_test_if(mark='skip', **kfilter):
    """Decorate test to skip/include certain parameters.

    Parameters:
        mark ({'skip', 'include', 'xfail'}):
            Whether the decorator should skip the test if it matches the filter
            conditions, include it only if it matches the conditions, or mark
            it as an expected failure.
        kfilter (kwargs):
            A keyword argument whose name should match one of the fixtures in
            the test. If the fixture returns a single value, the keyword must
            define a list of the fixture values to skip, include, or xfail.
            If the fixture returns an object, the value of the kwarg must be a
            dictionary of the object attributes to filter on. The ``mark`` is
            applied to all the attributes in the dictionary equally.

    Examples:
        If you want to only test for galaxies with bintype ``'STON'`` and
        template ``'MILES-THIN'`` you can do::

            @marvin_test_if(mark='include', galaxy=dict(bintype=['STON'], template=['MILES-THIN']))

        You can also mark all tests with ``data_origin=['file']`` as expected
        failure::

            @marvin_test_if(mark='xfails', data_origin=['file'])

        ``marvin_test_if`` decorators can be concatenated::

            @marvin_test_if(mark='xfails', data_origin=['file'])
            @marvin_test_if(mark='skip', galaxy=dict(bintype=['SPX']))

        will skip ``'SPX'`` bintypes and expect a failure on ``'file'``
        data_origins.

    """

    def _should_skip(filter_values, fixture_value, prop_name):
        ll = ', '.join(filter_values)
        if mark == 'skip' and fixture_value in filter_values:
            return pytest.skip('Skipping {0}={1!r}'.format(prop_name, ll))
        elif mark == 'include' and fixture_value not in filter_values:
            return pytest.skip('Skipping all {0} except {1!r}'.format(prop_name, ll))
        elif mark == 'xfail' and fixture_value in filter_values:
            return pytest.xfail('Expected failure if {0}={1!r}'.format(prop_name, ll))
        return False

    @decorator
    def decorated_function(ff, *args, **kwargs):

        ff_attr_names = inspect.getargspec(ff).args
        ff_attrs = {}
        for ii in range(len(args)):
            ff_attrs[ff_attr_names[ii]] = args[ii]

        assert mark in ['skip', 'include', 'xfail'], \
            'valid marks are \'skip\', \'include\', and \'xfail\''

        if len(kfilter) > 1:
            raise ValueError('marvin_test_if only accepts one filter condition.')

        fixture_to_filter, filter_attributes = list(kfilter.items())[0]

        if fixture_to_filter not in ff_attrs:
            return ff(*args, **kwargs)

        if not isinstance(filter_attributes, dict):
            _should_skip(filter_attributes, ff_attrs[fixture_to_filter], fixture_to_filter)
        else:
            for filter_attribute, filter_values in filter_attributes.items():
                fixture = ff_attrs[fixture_to_filter]
                if not hasattr(fixture, filter_attribute):
                    continue
                fixture_value = getattr(fixture, filter_attribute)
                if _should_skip(filter_values, fixture_value, filter_attribute):
                    break

        return ff(*args, **kwargs)
    return decorated_function


class marvin_test_if_class(object):
    """Decorate all tests in a class to run only for, or skip, certain parameters.

    See ``marvin_test_if``. This decorator is the equivalent for decorating
    classes isntead of functions.

    """

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __call__(self, decorated_class):
        for attr in inspect.getmembers(decorated_class, inspect.isfunction):
            # only decorate public functions
            if attr[0][0] != '_':
                setattr(decorated_class, attr[0],
                        marvin_test_if(*self.args,
                                       **self.kwargs)(getattr(decorated_class, attr[0])))
        return decorated_class


def skipIfNoDB(test):
    """Decorate a test to skip if DB ``session`` is ``None``."""
    @wraps(test)
    def wrapper(self, db, *args, **kwargs):
        if db.session is None:
            pytest.skip('Skip because no DB.')
        else:
            return test(self, db, *args, **kwargs)
    return wrapper



# Copyright 2015 SolidBuilds.com. All rights reserved.
#
# Authors: Ling Thio <ling.thio@gmail.com>

from __future__ import print_function
from sqlalchemy_boolean_search import parse_boolean_search
from .models import Record
import pytest


def add_records(db, records):
    for record in records:
        db.session.add(record)
    db.session.commit()


def delete_records(db, records):
    for record in records:
        db.session.delete(record)
    db.session.commit()


@pytest.mark.skip(reason="strings conditions no longer work the same way")
def test_strings(db):
    all_records = [
        Record(string='abc'),
        Record(string='abcx'),
        Record(string='xabc'),
        Record(string='xabcx'),
    ]
    add_records(db, all_records)

    expression = parse_boolean_search('string==abc')
    records = Record.query.filter(expression.filter(Record)).all()
    assert len(records) == 1
    for record in records:
        assert record.string == 'abc'

    expression = parse_boolean_search('string!=abc')
    records = Record.query.filter(expression.filter(Record)).all()
    assert len(records) == 3
    for record in records:
        assert record.string != 'abc'

    expression = parse_boolean_search('not string==abc')
    records = Record.query.filter(expression.filter(Record)).all()
    assert len(records) == 3
    for record in records:
        assert record.string != 'abc'

    expression = parse_boolean_search('string<xabc')
    records = Record.query.filter(expression.filter(Record)).all()
    assert len(records) == 2
    for record in records:
        assert record.string < 'xabc'

    expression = parse_boolean_search('string<=xabc')
    records = Record.query.filter(expression.filter(Record)).all()
    assert len(records) == 3
    for record in records:
        assert record.string <= 'xabc'

    expression = parse_boolean_search('string>xabc')
    records = Record.query.filter(expression.filter(Record)).all()
    assert len(records) == 1
    for record in records:
        assert record.string > 'xabc'

    expression = parse_boolean_search('string>=xabc')
    records = Record.query.filter(expression.filter(Record)).all()
    assert len(records) == 2
    for record in records:
        assert record.string >= 'xabc'

    expression = parse_boolean_search('string=abc')
    records = Record.query.filter(expression.filter(Record)).all()
    assert len(records) == 4
    for record in records:
        assert 'abc' in record.string

    expression = parse_boolean_search('string=*x and string=x*')
    records = Record.query.filter(expression.filter(Record)).all()
    assert len(records) == 1
    for record in records:
        assert record.string[0:1] == 'x' and record.string[-1:] == 'x'

    expression = parse_boolean_search('string=*x or string=x*')
    records = Record.query.filter(expression.filter(Record)).all()
    assert len(records) == 3
    for record in records:
        assert record.string[0:1] == 'x' or record.string[-1:] == 'x'

    delete_records(db, all_records)



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2016-10-19 17:36:00
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-08-02 14:17:05
#
# This is the Marvin setup
#

from setuptools import setup, find_packages

import os
import warnings
from get_version import generate_version_py

from astropy.utils.data import download_file

import argparse
import shutil
import sys


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def convert_md_to_rst(fp):
    try:
        import pypandoc
        output = pypandoc.convert_file(fp, 'rst')
        return output
    except ImportError:
        warnings.warn('cannot import pypandoc.', UserWarning)
        return open(fp).read()


def add_data_file(directory, data_files):
    extern_path = os.path.join(os.path.dirname(__file__), directory)
    for root, __, filenames in os.walk(extern_path):
        for filename in filenames:
            data_files.append(os.path.join('..', root.lstrip('python/'), filename))


def get_data_files(with_web=True):

    data_files = []

    # add_data_file('python/marvin/extern/', data_files)

    if with_web:
        add_data_file('python/marvin/web/configuration/', data_files)
        add_data_file('python/marvin/web/lib/', data_files)
        add_data_file('python/marvin/web/static/', data_files)
        add_data_file('python/marvin/web/templates/', data_files)
        add_data_file('python/marvin/web/uwsgi_conf_files/', data_files)

    # data_files.append('../marvin/db/dbconfig.ini')
    # data_files.append('../../requirements.txt')
    # data_files.append('../../README.md')
    # data_files.append('utils/plot/Linear_L_0-1.csv')

    return data_files


requirements_file = os.path.join(os.path.dirname(__file__), 'requirements.txt')
install_requires = [line.strip().replace('==', '>=') for line in open(requirements_file)
                    if not line.strip().startswith('#') and line.strip() != '']

NAME = 'sdss-marvin'
# do not use x.x.x-devx.  things complain.  instead use x.x.xdevx
VERSION = '2.1.4'
RELEASE = 'dev' not in VERSION
generate_version_py(NAME, VERSION, RELEASE)


def run(data_files, packages):

    setup(name=NAME,
          version=VERSION,
          license='BSD3',
          description='Toolsuite for dealing with the MaNGA dataset',
          long_description=convert_md_to_rst('README.md'),
          author='The Marvin Developers',
          author_email='havok2063@hotmail.com',
          keywords='marvin manga astronomy MaNGA',
          url='https://github.com/sdss/marvin',
          packages=packages,
          package_dir={'': 'python'},
          package_data={'': data_files},
          include_package_data=True,
          install_requires=install_requires,
          scripts=['bin/run_marvin', 'bin/check_marvin'],
          classifiers=[
              'Development Status :: 4 - Beta',
              'Environment :: Web Environment',
              'Framework :: Flask',
              'Intended Audience :: Science/Research',
              'License :: OSI Approved :: BSD License',
              'Natural Language :: English',
              'Operating System :: OS Independent',
              'Programming Language :: Python',
              'Programming Language :: Python :: 2.6',
              'Programming Language :: Python :: 2.7',
              'Programming Language :: Python :: 3.3',
              'Programming Language :: Python :: 3.4',
              'Topic :: Database :: Front-Ends',
              'Topic :: Documentation :: Sphinx',
              'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
              'Topic :: Scientific/Engineering :: Astronomy',
              'Topic :: Software Development :: Libraries :: Python Modules',
              'Topic :: Software Development :: User Interfaces'
          ],
          )


if __name__ == '__main__':

    # Custom parser to decide whether we include or not the web. By default we do.

    parser = argparse.ArgumentParser(prog=os.path.basename(sys.argv[0]))

    parser.add_argument('-w', '--noweb', dest='noweb', default=False,
                        action='store_true', help='Does not build the web.')

    # We use parse_known_args because we want to leave the remaining args for distutils
    args = parser.parse_known_args()[0]

    if args.noweb:
        packages = find_packages(where='python', exclude=['marvin.web*'])
    else:
        packages = find_packages(where='python')

    data_files = get_data_files(with_web=not args.noweb)

    maskbits_path = download_file('https://svn.sdss.org/public/repo/sdss/idlutils/'
                                  'trunk/data/sdss/sdssMaskbits.par')
    shutil.copy(maskbits_path, os.path.join(os.path.dirname(__file__),
                                            'python/marvin/data/',
                                            'sdssMaskbits.par'))

    # Now we remove all our custom arguments to make sure they don't interfere with distutils
    arguments = []
    for action in list(parser._get_optional_actions()):
        if '--help' not in action.option_strings:
            arguments += action.option_strings

    for arg in arguments:
        if arg in sys.argv:
            sys.argv.remove(arg)

    # Runs distutils
    run(data_files, packages)



#!/usr/bin/env python
# encoding: utf-8
#
# bin.py
#
# Created by José Sánchez-Gallego on 6 Nov 2016.


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import warnings

from marvin.core.exceptions import MarvinError, MarvinUserWarning, MarvinBreadCrumb
from marvin.tools.maps import Maps, _is_MPL4
from marvin.tools.modelcube import ModelCube
from marvin.tools.spaxel import Spaxel

breadcrumb = MarvinBreadCrumb()


class Bin(object):
    """A class to interface with a bin.

    This class represents a bin of spaxels in a MaNGA DAP
    :class:`~marvin.tools.maps.Maps` or
    :class:`~marvin.tools.modelcube.ModelCube`.
    By definition, a bin is identified by a ``binid`` and can correspond to a
    binned or unbinned MAPS, although in the latter case there is no formal
    difference between a bin and a spaxel.

    When a bin is instantiated, a list of its contituent
    :class:`~marvin.tools.spaxel.Spaxel` is created. ``Spaxel`` objects are
    created unloaded for efficiency, but they can be loaded by doing, for
    example for the first spaxel, ``bin.spaxels[0].load()``.

    MAPS properties and (if available) Model Cube spectra are loaded for the
    bin.

    Parameters:
        binid (int):
            The binid of the requested bin.
        extra_kwargs:
            Any other keyword parameter necessary to instantiate the
            :class:`~marvin.tools.maps.Maps`, the
            :class:`~marvin.tools.modelcube.ModelCube`, or the
            :class:`~marvin.tools.spaxel.Spaxel` related to this bin. Refer to
            the documentation of those classes for details on the available
            keywords.

    Attributes:
        spaxels (list):
            A list of :class:`~marvin.tools.spaxel.Spaxel` objects that form
            the bin. The spaxels are created unloaded.
        properties (dict):
            Same as the :class:`~marvin.tools.maps.Maps` properties for a
            given spaxel, but in this case for the bin.
        modelcube_attributes:
            All the attributes derived for a given spaxel from the
            :class:`~marvin.tools.modelcube.ModelCube`, but in this case for
            the bin. If a modelcube is not available or it is not instantiated,
            these attributes will be ``None``.

    """

    def __init__(self, binid, **kwargs):

        self.plateifu = None
        self.mangaid = None

        self._maps, self._modelcube = self._get_dap_objects(**kwargs)
        self.binid = binid

        # Drops some keyword that could make spaxel load fail.
        kwargs.pop('bintype', None)
        kwargs.pop('mode', None)

        # drop breadcrumb
        breadcrumb.drop(message='Initializing MarvinBin {0}'.format(self.__class__),
                        category=self.__class__)

        self._load_spaxels(**kwargs)
        self._load_data(**kwargs)

        self.release = self._maps.release

    def __repr__(self):
        return ('<Marvin Bin (binid={0}, n_spaxels={1}, bintype={2}, template_kin={3})>'
                .format(self.binid, len(self.spaxels), self._maps.bintype,
                        self._maps.template_kin))

    def _get_dap_objects(self, **kwargs):
        """Gets the Maps and ModelCube object."""

        try:
            kwargs_maps = kwargs.copy()
            kwargs_maps.pop('modelcube_filename', None)
            kwargs_maps['filename'] = kwargs_maps.pop('maps_filename', None)
            maps = Maps(**kwargs_maps)
        except MarvinError as ee:
            raise MarvinError('failed to open a Maps: {0}'.format(str(ee)))

        self.plateifu = maps.plateifu
        self.mangaid = maps.mangaid

        if _is_MPL4(maps._dapver):
            return maps, None

        # try:
        kwargs_modelcube = kwargs.copy()
        kwargs_modelcube.pop('maps_filename', None)
        kwargs_modelcube['filename'] = kwargs_modelcube.pop('modelcube_filename', None)
        kwargs_modelcube['plateifu'] = None

        # TODO: we need a check here to make sure that if we open both Maps and ModelCube
        # from files, their bintypes, templates, plate-ifu, etc are consistent.

        # TODO: this is not a very good implementation and probably has corner cases in which
        # it fails. It should be refactored.

        if kwargs_modelcube['filename'] is None:
            kwargs_modelcube['plateifu'] = self.plateifu
            kwargs_modelcube['bintype'] = maps.bintype
            kwargs_modelcube['template_kin'] = maps.template_kin
            kwargs_modelcube['template_pop'] = maps.template_pop
            kwargs_modelcube['release'] = maps._release

        modelcube = ModelCube(**kwargs_modelcube)

        # except Exception:
        #
        #     warnings.warn('cannot open a ModelCube for this combination of '
        #                   'parameters. Some fetures will not be available.', MarvinUserWarning)
        #     modelcube = False

        return maps, modelcube

    def _load_spaxels(self, **kwargs):
        """Creates a list of unloaded spaxels for this binid."""

        load_spaxels = kwargs.pop('load_spaxels', False)

        self.spaxel_coords = self._maps.get_bin_spaxels(self.binid, only_list=True)

        if len(self.spaxel_coords) == 0:
            raise MarvinError('there are no spaxels associated with binid={0}.'.format(self.binid))
        else:

            if 'plateifu' not in kwargs:
                kwargs['plateifu'] = self._maps.plateifu

            modelcube_for_spaxel = False if not self._modelcube else self._modelcube.get_unbinned()
            self.spaxels = [Spaxel(x=cc[0], y=cc[1], cube=True, maps=self._maps.get_unbinned(),
                                   modelcube=modelcube_for_spaxel,
                                   load=load_spaxels,
                                   **kwargs)
                            for cc in self.spaxel_coords]

    def _load_data(self, **kwargs):
        """Loads one of the spaxels to get the DAP properties for the binid."""

        assert len(self.spaxel_coords) > 0

        sample_coords = self.spaxel_coords[0]

        if 'plateifu' not in kwargs:
            kwargs['plateifu'] = self._maps.plateifu

        sample_spaxel = Spaxel(x=sample_coords[0], y=sample_coords[1],
                               cube=True, maps=self._maps, modelcube=self._modelcube,
                               load=True, allow_binned=True, **kwargs)

        self.specres = sample_spaxel.specres
        self.specresd = sample_spaxel.specresd
        self.spectrum = sample_spaxel.spectrum
        self.properties = sample_spaxel.properties

        self.model_flux = sample_spaxel.model_flux
        self.redcorr = sample_spaxel.redcorr
        self.model = sample_spaxel.model
        self.emline = sample_spaxel.emline
        self.emline_base = sample_spaxel.emline_base
        self.stellar_continuum = sample_spaxel.stellar_continuum

    def load_all(self):
        """Loads all the spaxels that for this bin."""

        for spaxel in self.spaxels:
            spaxel.load()

        return



#!/usr/bin/env python
# encoding: utf-8

# Licensed under a 3-clause BSD license.

from __future__ import print_function, division
import os
import numpy as np
import yaml
import yamlordereddictloader
from fuzzywuzzy import process

# Query Parameter Datamodel


def get_best_fuzzy(name, choices, cutoff=0):
    items = process.extractBests(name, choices, score_cutoff=cutoff)
    if items:
        scores = [s[1] for s in items]
        morethanone = sum(np.max(scores) == scores) > 1
        if morethanone:
            exact = [s for s in items if s[0].name.lower() == name.lower()]
            if exact:
                return exact[0]
            else:
                options = [s[0].name for s in items if s[1] == np.max(scores)]
                raise KeyError('{0} is too ambiguous.  Did you mean one of {1}?'.format(name, options))
        else:
            return items[0]
    else:
        return None


def get_params():
    bestpath = os.path.join(os.path.dirname(os.path.abspath(__file__)), '../../data', 'query_params_best.cfg')
    if os.path.isfile(bestpath):
        with open(bestpath, 'r') as stream:
            bestparams = yaml.load(stream, Loader=yamlordereddictloader.Loader)
        return bestparams
    else:
        return None


class ParameterGroupList(list):
    ''' ParameterGroup Object

    This object inherits from the Python list object. This
    represents a list of query ParameterGroups.

    '''

    def __init__(self, items):
        self.score = None
        if isinstance(items, list):
            list.__init__(self, items)
        elif isinstance(items, dict):
            paramgroups = [ParameterGroup(key, vals) for key, vals in items.items()]
            list.__init__(self, paramgroups)

    def list_groups(self):
        '''Returns a list of query groups.

        Returns:
            names (list):
                A string list of all the Query Group names
        '''
        return [group.name for group in self]

    def list_params(self, groups=None):
        '''Returns a list of parameters from all groups.

        Return a string list of the full parameter names.
        Default is all parameters across all groups.

        Parameters:
            groups (str|list):
                A string or list of strings representing the groups
                of parameters you wish to return

        Returns:
            params (list):
                A list of full parameter names
        '''
        if groups:
            groups = groups if isinstance(groups, list) else [groups]
            grouplist = [self[g] for g in groups]
            return [param.full for group in grouplist for param in group]
        else:
            return [param.full for group in self for param in group]

    def __eq__(self, name):
        item = get_best_fuzzy(name, self, cutoff=50)
        if item:
            self.score = item[1]
            return item[0]

    def __contains__(self, name):
        item = get_best_fuzzy(name, self, cutoff=50)
        if item:
            self.score = item[1]
            return item[0]
        else:
            return False

    def __getitem__(self, name):
        if isinstance(name, str):
            return self == name
        else:
            return list.__getitem__(self, name)


class ParameterGroup(list):
    ''' A Query Parameter Group Object

    Query parameters are grouped into specific categories
    for ease of use and navigation.  This object subclasses
    from the Python list object.

    Parameters:
        name (str):
            The name of the group

    '''
    def __init__(self, name, items):
        self.name = name
        self.score = None
        queryparams = [QueryParameter(**item) for item in items]
        list.__init__(self, queryparams)

    def __repr__(self):
        return ('<ParameterGroup name={0.name}, paramcount={1}>'.format(self, len(self)))

    def list_params(self, subset=None, display=None, short=None, full=None):
        ''' List the parameter names for a given group

        Lists the Query Parameters of the given group

        Parameters:
            subset (str|list):
                String list of a subset of parameters to return
            display (bool):
                Set to return the display names
            short (bool)
                Set to return the short names
            full (bool):
                Set to return the full names

        Returns:
            param (list):
                The list of parameter
        '''

        if subset:
            params = subset if isinstance(subset, list) else [subset]
            paramlist = [self[g] for g in params]
        else:
            paramlist = self

        if short:
            return [param.short for param in paramlist]
        elif display:
            return [param.display for param in paramlist]
        elif full:
            return [param.full for param in paramlist]
        else:
            return [param for param in paramlist]

    def __eq__(self, name):
        item = get_best_fuzzy(name, self, cutoff=25)
        if item:
            self.score = item[1]
            return item[0]

    def __contains__(self, name):
        item = get_best_fuzzy(name, self, cutoff=25)
        if item:
            self.score = item[1]
            return True
        else:
            return False

    def __getitem__(self, name):
        if isinstance(name, str):
            return self == name
        else:
            return list.__getitem__(self, name)


class QueryParameter(object):
    ''' A Query Parameter class

    An object representing a query parameter.  Provides access to
    different names for a given parameter.

    Parameters:
        full (str):
            The full naming syntax (table.name) used for all queries.  This name is recommended for full uniqueness.
        table (str):
            The name of the database table the parameter belongs to
        name (str):
            The name of the parameter in the database
        short (str):
            A shorthand name of the parameter
        display (str):
            A display name used for web and plotting purposes.
        dtype (str):
            The type of the parameter (e.g. string, integer, float)
    '''

    def __init__(self, full, table=None, name=None, short=None, display=None, dtype=None):
        self.full = full
        self.table = table
        self.name = name
        self.short = short
        self.display = display
        self.dtype = dtype
        self._joinedname = ', '.join([self.full, self.name, self.short, self.display])

    def __repr__(self):
        return ('<QueryParameter full={0.full}, name={0.name}, short={0.short}, display={0.display}>'.format(self))


bestparams = get_params()
query_params = ParameterGroupList(bestparams)






#!/usr/bin/env python3
# encoding: utf-8
#
# datamodel.py
#
# Created by José Sánchez-Gallego on 18 Sep 2016.


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from marvin import config


__all__ = ('MapsProperty', 'MapsPropertyList', 'dap_datamodel',
           'get_dap_datamodel', 'get_dap_maplist', 'get_default_mapset',
           'get_default_plot_params')


class MapsPropertyList(list):

    def __init__(self, items, version=None):

        list.__init__(self, items)
        self.version = version

    def get(self, value):
        """Returns the MapsProperty and channel matching a value."""

        for item in self:
            channels = [None] if not item.channels else item.channels
            for channel in channels:
                if value.lower() == item.fullname(channel):
                    return item, channel

        return None

    def list_names(self):
        """Returns a list of property names."""

        return [prop.name for prop in self]

    def __eq__(self, name):
        for item in self:
            if name.lower() == item.name:
                return item

    def __contains__(self, name):
        for item in self:
            if name.lower() == item.name:
                return True
        return False

    def __getitem__(self, name):
        if isinstance(name, str):
            return self == name
        else:
            return list.__getitem__(self, name)


class MapsProperty(object):
    """A class to represent a Maps property."""

    def __init__(self, name, ivar=False, mask=False, channels=None, unit=None, description=''):

        self.name = name.lower()
        self.ivar = ivar
        self.mask = mask
        self.channels = channels
        self.unit = unit
        self.description = description

        if self.channels:
            for ii in range(len(self.channels)):
                self.channels[ii] = self.channels[ii].lower()

    def __repr__(self):

        return ('<MapsProperty name={0.name}, ivar={0.ivar}, mask={0.mask}, n_channels={1}>'
                .format(self, len(self.channels) if self.channels else None))

    def fullname(self, channel=None, ext=None):

        if self.channels is None:
            if ext:
                return self.name + '_' + ext
            else:
                return self.name

        if channel is None:
            raise ValueError('this MapsProperty has multiple channels. Please, specify one.')

        if channel.lower() not in self.channels:
            raise ValueError('invalid channel.')

        if ext is not None:
            assert ext in ['ivar', 'mask'], 'ext must be one of ivar or mask.'
            return '{0}_{1}_{2}'.format(self.name, ext, channel.lower())

        return '{0}_{1}'.format(self.name, channel.lower())


MPL4_emline_channels = ['oiid_3728', 'hb_4862', 'oiii_4960', 'oiii_5008', 'oi_6302',
                        'oi_6365', 'nii_6549', 'ha_6564', 'nii_6585', 'sii_6718', 'sii_6732']

MPL4_specindex_channels = ['d4000', 'caii0p39', 'hdeltaa', 'cn1', 'cn2', 'ca4227', 'hgammaa',
                           'fe4668', 'hb', 'mgb', 'fe5270', 'fe5335', 'fe5406', 'nad', 'tio1',
                           'tio2', 'nai0p82', 'caii0p86a', 'caii0p86b', 'caii0p86c', 'mgi0p88',
                           'tio0p89', 'feh0p99']

MPL4_specindex_units = ['Angstrom', 'Angstrom', 'Angstrom', 'mag', 'mag', 'Angstrom', 'Angstrom',
                        'Angstrom', 'Angstrom', 'Angstrom', 'Angstrom', 'Angstrom', 'Angstrom',
                        'Angstrom', 'mag', 'mag', 'Angstrom', 'Angstrom', 'Angstrom', 'Angstrom',
                        'Angstrom', 'Angstrom', 'Angstrom']

MPL5_extra_channels = ['oii_3727', 'oii_3729', 'heps_3971', 'hdel_4102', 'hgam_4341', 'heii_4687',
                       'hei_5877', 'siii_8831', 'siii_9071', 'siii_9533']

default_version = '2.0.2'

dap_datamodel = {

    '1.1.1': MapsPropertyList([
        MapsProperty('emline_gflux', ivar=True, mask=True, channels=MPL4_emline_channels,
                     unit='1E-17 erg/s/cm^2/spaxel',
                     description='Fluxes of emission lines based on a single Gaussian fit.'),
        MapsProperty('emline_gvel', ivar=True, mask=True, channels=MPL4_emline_channels,
                     unit='km/s',
                     description='Doppler velocity shifts for emission lines relative to '
                                 'the NSA redshift based on a single Gaussian fit.'),
        MapsProperty('emline_gsigma', ivar=True, mask=True, channels=MPL4_emline_channels,
                     unit='km/s',
                     description='Velocity dispersions of emission lines based on a '
                                 'single Gaussian fit.'),
        MapsProperty('emline_instsigma', ivar=False, mask=False,
                     channels=MPL4_emline_channels,
                     unit='km/s',
                     description='Instrumental velocity dispersion at the line centroids '
                                 'for emission lines (based on a single Gaussian fit.'),
        MapsProperty('emline_ew', ivar=True, mask=True, channels=MPL4_emline_channels,
                     unit='Angstrom',
                     description='Equivalent widths for emission lines based on a '
                                 'single Gaussian fit.'),
        MapsProperty('emline_sflux', ivar=True, mask=True, channels=MPL4_emline_channels,
                     unit='1E-17 erg/s/cm^2/spaxel',
                     description='Fluxes for emission lines based on integrating the '
                                 'flux over a set of passbands.'),
        MapsProperty('stellar_vel', ivar=True, mask=True, channels=None,
                     unit='km/s',
                     description='Stellar velocity measurements.'),
        MapsProperty('stellar_sigma', ivar=True, mask=True, channels=None,
                     unit='km/s',
                     description='Stellar velocity dispersion measurements.'),
        MapsProperty('specindex', ivar=True, mask=True,
                     channels=MPL4_specindex_channels,
                     unit=None,
                     description='Measurements of spectral indices.'),
        MapsProperty('binid', ivar=False, mask=False, channels=None,
                     unit=None,
                     description='ID number for the bin for which the pixel value was '
                                 'calculated; bins are sorted by S/N.')],
        version='1.1.1'),

    '2.0.2': MapsPropertyList([
        MapsProperty('spx_skycoo', ivar=False, mask=False, channels=['on_sky_x', 'on_sky_y'],
                     unit='arcsec',
                     description='Offsets of each spaxel from the galaxy center.'),
        MapsProperty('spx_ellcoo', ivar=False, mask=False,
                     channels=['elliptical_radius', 'elliptical_azimuth'],
                     unit=['arcsec', 'degrees'],
                     description='Elliptical polar coordinates of each spaxel from '
                                 'the galaxy center.'),
        MapsProperty('spx_mflux', ivar=True, mask=False, channels=None,
                     unit='1E-17 erg/s/cm^2/ang/spaxel',
                     description='Mean flux in r-band (5600.1-6750.0 ang).'),
        MapsProperty('spx_snr', ivar=False, mask=False, channels=None,
                     unit=None,
                     description='r-band signal-to-noise ratio per pixel.'),
        MapsProperty('binid', ivar=False, mask=False, channels=None,
                     unit=None,
                     description='Numerical ID for spatial bins.'),
        MapsProperty('bin_lwskycoo', ivar=False, mask=False,
                     channels=['lum_weighted_on_sky_x', 'lum_weighted_on_sky_y'],
                     unit='arcsec',
                     description='Light-weighted offset of each bin from the galaxy center.'),
        MapsProperty('bin_lwellcoo', ivar=False, mask=False,
                     channels=['lum_weighted_elliptical_radius',
                               'lum_weighted_elliptical_azimuth'],
                     unit=['arcsec', 'degrees'],
                     description='light-weighted elliptical polar coordinates of each bin '
                                 'from the galaxy center.'),
        MapsProperty('bin_area', ivar=False, mask=False, channels=None,
                     unit='arcsec^2',
                     description='Area of each bin.'),
        MapsProperty('bin_farea', ivar=False, mask=False, channels=None,
                     unit=None,
                     description='Fractional area that the bin covers for the expected bin '
                                 'shape (only relevant for radial binning).'),
        MapsProperty('bin_mflux', ivar=True, mask=True, channels=None,
                     unit='1E-17 erg/s/cm^2/ang/spaxel',
                     description='Mean flux in the r-band for the binned spectra.'),
        MapsProperty('bin_snr', ivar=False, mask=False, channels=None,
                     unit=None,
                     description='r-band signal-to-noise ratio per pixel in the binned spectra.'),
        MapsProperty('stellar_vel', ivar=True, mask=True, channels=None,
                     unit='km/s',
                     description='Stellar velocity relative to NSA redshift.'),
        MapsProperty('stellar_sigma', ivar=True, mask=True, channels=None,
                     unit='km/s',
                     description='Stellar velocity dispersion (must be corrected using '
                                 'STELLAR_SIGMACORR)'),
        MapsProperty('stellar_sigmacorr', ivar=False, mask=False, channels=None,
                     unit='km/s',
                     description='Quadrature correction for STELLAR_SIGMA to obtain the '
                                 'astrophysical velocity dispersion.)'),
        MapsProperty('stellar_cont_fresid', ivar=False, mask=False,
                     channels=['68th_percentile', '99th_percentile'],
                     unit=None,
                     description='68%% and 99%% growth of the fractional residuals between '
                                 'the model and data'),
        MapsProperty('stellar_cont_rchi2', ivar=False, mask=False, channels=None,
                     unit=None,
                     description='Reduced chi-square of the stellar continuum fit.'),
        MapsProperty('emline_sflux', ivar=True, mask=True,
                     channels=MPL4_emline_channels + MPL5_extra_channels,
                     unit='1E-17 erg/s/cm^2/spaxel',
                     description='Non-parametric summed flux for emission lines.'),
        MapsProperty('emline_sew', ivar=True, mask=True,
                     channels=MPL4_emline_channels + MPL5_extra_channels,
                     unit='Angstrom',
                     description='Emission line non-parametric equivalent widths measurements.'),
        MapsProperty('emline_gflux', ivar=True, mask=True,
                     channels=MPL4_emline_channels + MPL5_extra_channels,
                     unit='1E-17 erg/s/cm^2/spaxel',
                     description='Gaussian profile integrated flux for emission lines.'),
        MapsProperty('emline_gvel', ivar=True, mask=True,
                     channels=MPL4_emline_channels + MPL5_extra_channels,
                     unit='km/s',
                     description='Gaussian profile velocity for emission lines.'),
        MapsProperty('emline_gsigma', ivar=True, mask=True,
                     channels=MPL4_emline_channels + MPL5_extra_channels,
                     unit='km/s',
                     description='Gaussian profile velocity dispersion for emission lines; '
                                 'must be corrected using EMLINE_INSTSIGMA'),
        MapsProperty('emline_instsigma', ivar=False, mask=False,
                     channels=MPL4_emline_channels + MPL5_extra_channels,
                     unit='km/s',
                     description='Instrumental dispersion at the fitted line center.'),
        MapsProperty('specindex', ivar=True, mask=True,
                     channels=['d4000', 'dn4000'],
                     unit=None,
                     description='Measurements of spectral indices.'),
        MapsProperty('specindex_corr', ivar=False, mask=False,
                     channels=['d4000', 'dn4000'],
                     unit=None,
                     description='Velocity dispersion corrections for the '
                                 'spectral index measurements '
                                 '(can be ignored for D4000, Dn4000).')],
        version='2.0.2')}


def get_dap_datamodel(dapver=None):
    """Returns the correct DAP datamodel for dapver."""

    if not dapver:
        __, dapver = config.lookUpVersions(config.release)

    if dapver not in dap_datamodel:
        return dap_datamodel[default_version]
    else:
        return dap_datamodel[dapver]


def get_dap_maplist(dapver=None, web=None):
    ''' Returns a list of all possible maps for dapver '''

    dapdm = get_dap_datamodel(dapver)
    daplist = []

    for p in dapdm:
        if p.channels:
            if web:
                daplist.extend(['{0}:{1}'.format(p.name, c) for c in p.channels])
            else:
                daplist.extend(['{0}_{1}'.format(p.name, c) for c in p.channels])
        else:
            daplist.append(p.name)

    return daplist


def get_default_mapset(dapver=None):
    ''' Returns a list of six default maps for display '''

    dapdefaults = {
        # 6 defaults
        # '1.1.1': ['emline_gflux:oiid_3728', 'emline_gflux:hb_4862', 'emline_gflux:oiii_5008',
        #           'emline_gflux:ha_6564', 'emline_gflux:nii_6585', 'emline_gflux:sii_6718'],
        # '2.0.2': ['emline_gflux:oiid_3728', 'emline_gflux:hb_4862', 'emline_gflux:oiii_5008',
        #           'emline_gflux:ha_6564', 'emline_gflux:nii_6585', 'emline_gflux:sii_6718']
        # 3 defaults
        '1.1.1': ['stellar_vel', 'emline_gflux:ha_6564', 'specindex:d4000'],
        '2.0.2': ['stellar_vel', 'emline_gflux:ha_6564', 'specindex:d4000']
    }

    return dapdefaults[dapver] if dapver in dapdefaults else []


def get_default_plot_params(dapver=None):
    """Returns default map plotting parameters."""
    bitmasks = {'1.1.1': {'badData': {'doNotUse': 0}},
                '2.0.2': {'nocov': 0,
                          'badData': {'unreliable': 5,
                                      'doNotUse': 30}
                          }
                }
    plot_defaults = {
        '1.1.1': {'default': {'bitmasks': bitmasks['1.1.1'],
                              'cmap': 'linearlab',
                              'percentile_clip': [5, 95],
                              'symmetric': False,
                              'snr_min': 1},
                  'vel': {'bitmasks': bitmasks['1.1.1'],
                          'cmap': 'RdBu_r',
                          'percentile_clip': [10, 90],
                          'symmetric': True,
                          'snr_min': None},
                  'sigma': {'bitmasks': bitmasks['1.1.1'],
                            'cmap': 'inferno',
                            'percentile_clip': [10, 90],
                            'symmetric': False,
                            'snr_min': 1}},
        '2.0.2': {'default': {'bitmasks': bitmasks['2.0.2'],
                              'cmap': 'linearlab',
                              'percentile_clip': [5, 95],
                              'symmetric': False,
                              'snr_min': 1},
                  'vel': {'bitmasks': bitmasks['2.0.2'],
                          'cmap': 'RdBu_r',
                          'percentile_clip': [10, 90],
                          'symmetric': True,
                          'snr_min': None},
                  'sigma': {'bitmasks': bitmasks['2.0.2'],
                            'cmap': 'inferno',
                            'percentile_clip': [10, 90],
                            'symmetric': False,
                            'snr_min': 1}}
    }
    return plot_defaults[dapver] if dapver in plot_defaults else {}



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-05-07 13:48:11
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-07-28 11:03:45

from __future__ import print_function, division, absolute_import
from marvin.tests.web.conftest import Page
from marvin import config
from marvin.web import create_app
from marvin.api.base import arg_validate as av
from marvin.web.settings import TestConfig, CustomConfig
from marvin.web.extensions import limiter
import pytest
import os
import six

# releases = ['MPL-5']


# @pytest.fixture(scope='session', params=releases)
# def release(request):
#     return request.param


# @pytest.fixture(scope='session')
# def drpver(release):
#     drpver, dapver = config.lookUpVersions(release)
#     return drpver


# @pytest.fixture(scope='session')
# def dapver(release):
#     drpver, dapver = config.lookUpVersions(release)
#     return dapver


# @pytest.fixture(scope='session')
# def mode():
#     return config.mode


# @pytest.fixture(scope='session')
# def app():
#     app = create_app(debug=True, local=True, use_profiler=False)
#     app.config['TESTING'] = True
#     app.config['WTF_CSRF_ENABLED'] = False
#     app.config['PRESERVE_CONTEXT_ON_EXCEPTION'] = False
#     return app

@pytest.fixture(scope='session')
def app():
    object_config = type('Config', (TestConfig, CustomConfig), dict())
    app = create_app(debug=True, local=True, object_config=object_config)
    limiter.enabled = False
    return app


@pytest.fixture(scope='session')
def init_api(urlmap):
    config.urlmap = urlmap
    config.forceDbOn()


class ApiPage(Page):

    def __init__(self, client, blue, endpoint):
        super(ApiPage, self).__init__(client, blue, endpoint)
        self.api_base_success = dict(status=1, error=None, traceback=None)

    def assert_success(self, expdata=None, keys=None, issubset=None):
        self.assert200(message='response status should be 200 for ok and not {0}'.format(self.response.status_code))
        assert self.json['status'] == 1
        self.assert_dict_contains_subset(self.api_base_success, self.json)
        if isinstance(expdata, str):
            assert isinstance(self.json['data'], six.string_types), 'response data should be a string'
            assert expdata in self.json['data']
        elif isinstance(expdata, dict):
            assert isinstance(self.json['data'], dict), 'response data should be a dict'
            if keys:
                assert set(expdata.keys()) == set(self.json['data'].keys())
            else:
                #assert expdata.items() <= self.json['data'].items()
                self.assert_dict_contains_subset(expdata, self.json['data'])
        elif isinstance(expdata, list):
            assert isinstance(self.json['data'], list), 'response data should be a list'
            if issubset:
                subset = all(row in self.json['data'] for row in expdata)
                assert subset is True, 'one should be subset of the other'
            else:
                assert expdata == self.json['data'], 'two lists should be the same'

    def route_no_valid_params(self, url, noparam, reqtype='get', params=None, errmsg=None):
        self.load_page(reqtype, url, params=params)
        self.assert422(message='response status should be 422 for invalid params')
        assert 'validation_errors' in self.json.keys()
        noparam = [noparam] if not isinstance(noparam, list) else noparam
        errmsg = [errmsg] if not isinstance(errmsg, list) else errmsg
        invalid = {p: errmsg for p in noparam}
        self.assert_dict_contains_subset(invalid, self.json['validation_errors'])


@pytest.fixture()
def page(client, request, init_api):
    blue, endpoint = request.param
    page = ApiPage(client, 'api', endpoint)
    yield page


@pytest.fixture()
def params(galaxy):
    return {'release': galaxy.release}



#!/usr/bin/env python
# encoding: utf-8
#
# test_maps.py
#
# Created by José Sánchez-Gallego on 22 Jun 2016.


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import pytest

import astropy.io.fits

import marvin
from marvin.tools.maps import Maps
import marvin.tools.spaxel
from marvin.core.exceptions import MarvinError
from marvin.tests import marvin_test_if


def _assert_maps(maps, galaxy):
    """Basic checks for a Maps object."""

    assert maps is not None
    assert maps.plateifu == galaxy.plateifu
    assert maps.mangaid == galaxy.mangaid
    assert maps.wcs is not None
    assert maps.bintype == galaxy.bintype

    assert len(maps.shape) == len(galaxy.shape)
    for ii in range(len(maps.shape)):
        assert maps.shape[ii] == galaxy.shape[ii]


class TestMaps(object):

    def _get_maps_kwargs(self, galaxy, data_origin):

        if data_origin == 'file':
            maps_kwargs = dict(filename=galaxy.mapspath)
        else:
            maps_kwargs = dict(plateifu=galaxy.plateifu, release=galaxy.release,
                               bintype=galaxy.bintype, template_kin=galaxy.template,
                               mode='local' if data_origin == 'db' else 'remote')

        return maps_kwargs

    def test_load(self, galaxy, exporigin):

        maps = Maps(**self._get_maps_kwargs(galaxy, exporigin))

        _assert_maps(maps, galaxy)

        assert maps.data_origin == exporigin

        if exporigin == 'file':
            assert isinstance(maps.data, astropy.io.fits.HDUList)
        elif exporigin == 'db':
            assert isinstance(maps.data, marvin.marvindb.dapdb.File)

        assert maps.cube is not None
        assert maps.cube.plateifu == galaxy.plateifu
        assert maps.cube.mangaid == galaxy.mangaid

    @pytest.mark.parametrize('monkeyconfig', [('release', 'MPL-5')], indirect=True)
    def test_load_mpl4_global_mpl5(self, galaxy, monkeyconfig, data_origin):

        assert marvin.config.release == 'MPL-5'
        maps = Maps(**self._get_maps_kwargs(galaxy, data_origin))

        assert maps._release == galaxy.release
        assert maps._drpver == galaxy.drpver
        assert maps._dapver == galaxy.dapver

    def test_get_spaxel(self, galaxy, data_origin):

        maps = Maps(**self._get_maps_kwargs(galaxy, data_origin))
        spaxel = maps.getSpaxel(x=15, y=8, xyorig='lower')

        assert isinstance(spaxel, marvin.tools.spaxel.Spaxel)
        assert spaxel.spectrum is not None
        assert len(spaxel.properties.keys()) > 0

        expected = galaxy.stellar_vel_ivar_x15_y8_lower[galaxy.release][galaxy.template]
        assert spaxel.properties['stellar_vel'].ivar == pytest.approx(expected, abs=1e-6)

    def test_get_spaxel_test2(self, galaxy, data_origin):

        maps = Maps(**self._get_maps_kwargs(galaxy, data_origin))
        spaxel = maps.getSpaxel(x=5, y=5)

        assert isinstance(spaxel, marvin.tools.spaxel.Spaxel)
        assert spaxel.spectrum is not None
        assert len(spaxel.properties.keys()) > 0

    def test_get_spaxel_no_db(self, galaxy, exporigin):
        """Tests getting an spaxel if there is no DB."""

        maps = Maps(**self._get_maps_kwargs(galaxy, exporigin))
        spaxel = maps.getSpaxel(x=5, y=5)

        assert spaxel.maps.data_origin == exporigin

        assert isinstance(spaxel, marvin.tools.spaxel.Spaxel)
        assert spaxel.spectrum is not None
        assert len(spaxel.properties.keys()) > 0

    def test_maps_redshift(self, maps, galaxy):
        redshift = maps.nsa.redshift if maps.release == 'MPL-4' and maps.data_origin == 'file' else maps.nsa.z
        assert pytest.approx(redshift, galaxy.redshift)

    def test_release(self, galaxy):
        maps = Maps(plateifu=galaxy.plateifu)
        assert maps.release == galaxy.release

    def test_set_release_fails(self, galaxy):
        maps = Maps(plateifu=galaxy.plateifu)
        with pytest.raises(MarvinError) as ee:
            maps.release = 'a'
            assert 'the release cannot be changed' in str(ee.exception)



from decimal import Decimal

import six
import sqlalchemy as sa
from wtforms_components import SelectField

from tests import ModelFormTestCase


class MultiDict(dict):
    def getlist(self, key):
        return [self[key]]


class TestSelectFieldDefaultValue(ModelFormTestCase):
    def test_option_selected_by_field_default_value(self):
        choices = [(u'1', '1'), (u'2', '2')]
        self.init(
            type_=sa.Integer,
            default=u'1',
            info={'choices': choices}
        )
        form = self.form_class(MultiDict({'test_column': '2'}))
        assert (
            '<option selected value="2">2</option>' in
            str(form.test_column)
        )


class TestSelectFieldCoerce(ModelFormTestCase):
    def test_integer_coerces_values_to_integers(self):
        choices = [(u'1', '1'), (u'2', '2')]
        self.init(type_=sa.Integer, info={'choices': choices})
        form = self.form_class(MultiDict({'test_column': '2'}))
        assert form.test_column.data is 2

    def test_nullable_integer_coerces_values_to_integers(self):
        choices = [(u'1', '1'), (u'2', '2')]
        self.init(type_=sa.Integer, nullable=True, info={'choices': choices})
        form = self.form_class(MultiDict({'test_column': '2'}))
        assert form.test_column.data is 2

    def test_integer_coerces_empty_strings_to_nulls(self):
        choices = [(u'1', '1'), (u'2', '2')]
        self.init(type_=sa.Integer, info={'choices': choices})
        form = self.form_class(MultiDict({'test_column': ''}))
        assert form.test_column.data is None

    def test_big_integer_coerces_values_to_integers(self):
        choices = [(u'1', '1'), (u'2', '2')]
        self.init(type_=sa.BigInteger, info={'choices': choices})
        self.assert_type('test_column', SelectField)
        form = self.form_class(MultiDict({'test_column': '2'}))
        assert form.test_column.data is 2

    def test_small_integer_coerces_values_to_integers(self):
        choices = [(u'1', '1'), (u'2', '2')]
        self.init(type_=sa.SmallInteger, info={'choices': choices})
        form = self.form_class(MultiDict({'test_column': '2'}))
        assert form.test_column.data is 2

    def test_numeric_coerces_values_to_decimals(self):
        choices = [(u'1.0', '1.0'), (u'2.0', '2.0')]
        self.init(type_=sa.Numeric, info={'choices': choices})
        form = self.form_class(MultiDict({'test_column': '2.0'}))
        assert form.test_column.data == Decimal('2.0')

    def test_float_coerces_values_to_floats(self):
        choices = [(u'1.0', '1.0'), (u'2.0', '2.0')]
        self.init(type_=sa.Float, info={'choices': choices})
        form = self.form_class(MultiDict({'test_column': '2.0'}))
        assert form.test_column.data == 2.0

    def test_unicode_coerces_values_to_unicode_strings(self):
        choices = [('1.0', '1.0'), ('2.0', '2.0')]
        self.init(type_=sa.Unicode(255), info={'choices': choices})
        form = self.form_class(MultiDict({'test_column': '2.0'}))
        assert form.test_column.data == u'2.0'
        assert isinstance(form.test_column.data, six.text_type)

    def test_unicode_text_coerces_values_to_unicode_strings(self):
        choices = [('1.0', '1.0'), ('2.0', '2.0')]
        self.init(type_=sa.UnicodeText, info={'choices': choices})
        form = self.form_class(MultiDict({'test_column': '2.0'}))
        assert form.test_column.data == u'2.0'
        assert isinstance(form.test_column.data, six.text_type)



# Copyright 2015 SolidBuilds.com. All rights reserved.
#
# Authors: Ling Thio <ling.thio@gmail.com>

from __future__ import print_function
from sqlalchemy_boolean_search import parse_boolean_search, BooleanSearchException
from .models import Record


def add_records(db, records):
    for record in records:
        db.session.add(record)
    db.session.commit()


def delete_records(db, records):
    for record in records:
        db.session.delete(record)
    db.session.commit()


def test_others(db):
    # Create records
    all_records = [
        Record(integer=1),
        Record(float=1.0),
    ]
    add_records(db, all_records)

    # Test integer
    expression = parse_boolean_search('integer==1')
    records = Record.query.filter(expression.filter(Record)).all()
    assert len(records) == 1
    for record in records:
        assert record.integer == 1

    # Test float with float
    expression = parse_boolean_search('float==1.0')
    records = Record.query.filter(expression.filter(Record)).all()
    assert len(records) == 1
    for record in records:
        assert record.float == 1.0

    # Test float with integer
    expression = parse_boolean_search('float==1')
    records = Record.query.filter(expression.filter(Record)).all()
    assert len(records) == 1
    for record in records:
        assert record.float == 1.0

    # Delete records
    delete_records(db, all_records)



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-02-12 20:46:42
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-07-13 16:41:23

from __future__ import print_function, division, absolute_import
import pytest
from marvin.tests.web.conftest import Page
from marvin import marvindb
from flask import url_for


@pytest.mark.parametrize('page', [('index_page', 'Marvin:index')], ids=['index'], indirect=True)
class TestIndexPage(object):

    def test_assert_index_template_used(self, page, get_templates):
        page.load_page('get', page.url)
        assert '' == page.data
        template, context = get_templates[0]
        assert 'index.html' == template.name


@pytest.mark.parametrize('page', [('index_page', 'Marvin:database')], ids=['database'], indirect=True)
class TestDb(object):

    def test_db_works(self, page, release):
        page.load_page('get', page.url, params={'release': release})
        data = {'plate': 7443}
        page.assert_webjson_success(data)

    def test_db_post_fails(self, page, release):
        page.load_page('post', page.url, params={'release': release})
        page.assert405('allowed method should be get')


@pytest.mark.parametrize('page', [('index_page', 'selectmpl')], ids=['selectmpl'], indirect=True)
class TestSelectMPL(object):

    def test_select_mpl(self, page, release, drpver, dapver):
        page.load_page('post', page.url, params={'release': release})
        data = {'current_release': release, 'current_drpver': drpver, 'current_dapver': dapver}
        page.assert_webjson_success(data)
        self._release_in_session(page, data)

    def _release_in_session(self, page, data):
        with page.client.session_transaction() as sess:
            sess['release'] = data['current_release']
            sess['drpver'] = data['current_drpver']
            sess['dapver'] = data['current_dapver']


@pytest.mark.parametrize('page', [('index_page', 'getgalidlist')], ids=['getgalid'], indirect=True)
class TestGetGalIdList(object):

    def test_getgalid_success(self, page, release):
        page.load_page('post', page.url, params={'release': release})
        data = ['8485', '8485-1901', '7443', '7443-12701', '1-209232', '12-98126']
        page.assert200(message='response status should be 200 for ok')
        page.assertListIn(data, page.json)

    def test_getgalid_fail(self, page, release):
        marvindb.datadb = None
        page.load_page('post', page.url, params={'release': release})
        data = ['']
        page.assert200(message='response status should be 200 for ok')
        assert data == page.json


@pytest.mark.parametrize('page', [('index_page', 'galidselect')], ids=['galidselect'], indirect=True)
@pytest.mark.parametrize('name, id, galid', [('plate', 'plate', 8485),
                                             ('galaxy', 'plateifu', '8485-1901'),
                                             ('galaxy', 'mangaid', '1-209232'),
                                             ('main', None, None)])
class TestGalIdSelect(object):

    @pytest.fixture(autouse=True)
    def get_url(self, name, galid):
        if name == 'plate':
            return url_for('plate_page.Plate:get', plateid=galid)
        elif name == 'plateifu':
            return url_for('galaxy_page.Galaxy:get', galid=galid)
        elif name == 'mangaid':
            return url_for('galaxy_page.Galaxy:get', galid=galid)
        elif name is None:
            return url_for('index_page.Marvin:index')

    def test_get_galid(self, page, release, name, id, galid):
        data = {'galid': galid, 'release': release}
        page.load_page('get', page.url, params=data)
        redirect_url = self.get_url(id, galid)
        if id:
            page.assert_redirects(redirect_url, 'page should be redirected to {0} page'.format(name))
        else:
            page.assert422(message='response should be 422 for no name input')


@pytest.mark.parametrize('page', [('index_page', 'login')], ids=['login'], indirect=True)
@pytest.mark.parametrize('data, exp',
                         [({'username': '', 'password': ''}, {'ready': False, 'status': -1, 'message': ''}),
                          ({'username': 'sdss', 'password': 'password'}, {'ready': False, 'status': 0, 'message': 'Failed login for sdss. Please retry.', 'membername': 'Unknown user'}),
                          ({'username': 'bac29', 'password': 'password'}, {'ready': False, 'status': 0, 'message': 'Failed login for bac29. Username unrecognized.', 'membername': 'Unknown user'}),
                          pytest.mark.xfail(reason="not real login")(({'username': 'sdss', 'password': 'password'}, {'ready': True, 'status': 1, 'message': 'Logged in as sdss. ', 'membername': 'SDSS User'}))],
                         ids=['no_input', 'wrong_pass', 'wrong_user', 'success'])
class TestLogin(object):

    def test_login(self, inspection, page, release, data, exp):
        data['release'] = release
        page.load_page('post', page.url, params=data)
        page.assert200('response status should be 200 for ok')
        assert exp['status'] == page.json['result']['status']
        assert exp['message'] == page.json['result']['message']
        if 'membername' in exp:
            assert exp['membername'] == page.json['result']['membername']





import collections
import os
import warnings

import numpy as np
import PIL
from scipy.interpolate import griddata

from astropy import wcs
from astropy import table

import marvin
from marvin import log
from marvin.utils.dap.datamodel import get_default_plot_params
from marvin.core.exceptions import MarvinError, MarvinUserWarning
from brain.core.exceptions import BrainError

try:
    from sdss_access import RsyncAccess, AccessError
except ImportError as e:
    RsyncAccess = None

try:
    from sdss_access.path import Path
except ImportError as e:
    Path = None


# General utilities
__all__ = ('convertCoords', 'parseIdentifier', 'mangaid2plateifu', 'findClosestVector',
           'getWCSFromPng', 'convertImgCoords', 'getSpaxelXY',
           'downloadList', 'getSpaxel', 'get_drpall_row', 'getDefaultMapPath',
           'getDapRedux', 'get_nsa_data', '_check_file_parameters', 'get_plot_params')

drpTable = {}


def getSpaxel(cube=True, maps=True, modelcube=True,
              x=None, y=None, ra=None, dec=None, xyorig=None):
    """Returns the |spaxel| matching certain coordinates.

    The coordinates of the spaxel to return can be input as ``x, y`` pixels
    relative to``xyorig`` in the cube, or as ``ra, dec`` celestial
    coordinates.

    This function is intended to be called by
    :func:`~marvin.tools.cube.Cube.getSpaxel` or
    :func:`~marvin.tools.maps.Maps.getSpaxel`, and provides shared code for
    both of them.

    Parameters:
        cube (:class:`~marvin.tools.cube.Cube` or None or bool)
            A :class:`~marvin.tools.cube.Cube` object with the DRP cube
            data from which the spaxel spectrum will be extracted. If None,
            the |spaxel| object(s) returned won't contain spectral information.
        maps (:class:`~marvin.tools.maps.Maps` or None or bool)
            As ``cube`` but for the :class:`~marvin.tools.maps.Maps`
            object representing the DAP maps entity. If None, the |spaxel|
            will be returned without DAP information.
        modelcube (:class:`~marvin.tools.modelcube.ModelCube` or None or bool)
            As ``cube`` but for the :class:`~marvin.tools.modelcube.ModelCube`
            object representing the DAP modelcube entity. If None, the |spaxel|
            will be returned without model information.
        x,y (int or array):
            The spaxel coordinates relative to ``xyorig``. If ``x`` is an
            array of coordinates, the size of ``x`` must much that of
            ``y``.
        ra,dec (float or array):
            The coordinates of the spaxel to return. The closest spaxel to
            those coordinates will be returned. If ``ra`` is an array of
            coordinates, the size of ``ra`` must much that of ``dec``.
        xyorig ({'center', 'lower'}):
            The reference point from which ``x`` and ``y`` are measured.
            Valid values are ``'center'``, for the centre of the
            spatial dimensions of the cube, or ``'lower'`` for the
            lower-left corner. This keyword is ignored if ``ra`` and
            ``dec`` are defined. ``xyorig`` defaults to
            ``marvin.config.xyorig.``

    Returns:
        spaxels (list):
            The |spaxel| objects for this cube/maps corresponding to the input
            coordinates. The length of the list is equal to the number
            of input coordinates.

    .. |spaxel| replace:: :class:`~marvin.tools.spaxel.Spaxel`

    """

    # TODO: for now let's put these imports here, but we should fix the
    # circular imports soon.
    import marvin.tools.cube
    import marvin.tools.maps
    import marvin.tools.modelcube
    import marvin.tools.spaxel

    # Checks that the cube and maps data are correct
    assert cube or maps or modelcube, \
        'Either cube, maps, or modelcube needs to be specified.'

    assert isinstance(cube, (marvin.tools.cube.Cube, bool)), \
        'cube is not an instance of Cube or a boolean'

    assert isinstance(maps, (marvin.tools.maps.Maps, bool)), \
        'maps is not an instance of Maps or a boolean'

    assert isinstance(modelcube, (marvin.tools.modelcube.ModelCube, bool)), \
        'modelcube is not an instance of ModelCube or a boolean'

    # Checks that we have the correct set of inputs.
    if x is not None or y is not None:
        assert ra is None and dec is None, 'Either use (x, y) or (ra, dec)'
        assert x is not None and y is not None, 'Specify both x and y'

        inputMode = 'pix'
        isScalar = np.isscalar(x)
        x = np.atleast_1d(x)
        y = np.atleast_1d(y)
        coords = np.array([x, y], np.float).T

    elif ra is not None or dec is not None:
        assert x is None and y is None, 'Either use (x, y) or (ra, dec)'
        assert ra is not None and dec is not None, 'Specify both ra and dec'

        inputMode = 'sky'
        isScalar = np.isscalar(ra)
        ra = np.atleast_1d(ra)
        dec = np.atleast_1d(dec)
        coords = np.array([ra, dec], np.float).T

    else:
        raise ValueError('You need to specify either (x, y) or (ra, dec)')

    if not xyorig:
        xyorig = marvin.config.xyorig

    if isinstance(maps, marvin.tools.maps.Maps):
        ww = maps.wcs if inputMode == 'sky' else None
        cube_shape = maps.shape
        plateifu = maps.plateifu
    elif isinstance(cube, marvin.tools.cube.Cube):
        ww = cube.wcs if inputMode == 'sky' else None
        cube_shape = cube.shape
        plateifu = cube.plateifu
    elif isinstance(modelcube, marvin.tools.modelcube.ModelCube):
        ww = modelcube.wcs if inputMode == 'sky' else None
        cube_shape = modelcube.shape
        plateifu = modelcube.plateifu

    iCube, jCube = zip(convertCoords(coords, wcs=ww, shape=cube_shape,
                                     mode=inputMode, xyorig=xyorig).T)

    _spaxels = []
    for ii in range(len(iCube[0])):
        _spaxels.append(
            marvin.tools.spaxel.Spaxel(x=jCube[0][ii], y=iCube[0][ii],
                                       cube=cube, maps=maps, modelcube=modelcube))

    if len(_spaxels) == 1 and isScalar:
        return _spaxels[0]
    else:
        return _spaxels


def convertCoords(coords, mode='sky', wcs=None, xyorig='center', shape=None):
    """Converts input coordinates to array indices.

    Converts input positions in x, y or RA, Dec coordinates to array indices
    (in Numpy style) or spaxel extraction. In case of pixel coordinates, the
    origin of reference (either the center of the cube or the lower left
    corner) can be specified via ``xyorig``.

    If ``shape`` is defined (mandatory for ``mode='pix'``, optional for
    ``mode='sky'``) and one or more of the resulting indices are outside the
    size of the input shape, an error is raised.

    This functions is mostly intended for internal use.

    Parameters:
        coords (array):
            The input coordinates, as an array of shape Nx2.
        mode ({'sky', 'pix'}:
            The type of input coordinates, either `'sky'` for celestial
            coordinates (in the format defined in the WCS header information),
            or `'pix'` for pixel coordinates.
        wcs (None or ``astropy.wcs.WCS`` object):
            If ``mode='sky'``, the WCS solution from which the cube coordinates
            can be derived.
        xyorig (str):
            If ``mode='pix'``, the reference point from which the coordinates
            are measured. Valid values are ``'center'``, for the centre of the
            spatial dimensions of the cube, or ``'lower'`` for the lower-left
            corner.
        shape (None or array):
            If ``mode='pix'``, the shape of the spatial dimensions of the cube,
            so that the central position can be calculated.

    Returns:
        result (Numpy array):
            An array with the same shape as ``coords``, containing the cube
            index positions for the input coordinates, in Numpy style (i.e.,
            the first element being the row and the second the column).

    """

    coords = np.atleast_2d(coords)
    assert coords.shape[1] == 2, 'coordinates must be an array Nx2'

    if mode == 'sky':
        assert wcs, 'if mode==sky, wcs must be defined.'
        coordsSpec = np.ones((coords.shape[0], 3), np.float32)
        coordsSpec[:, :-1] = coords
        cubeCoords = wcs.wcs_world2pix(coordsSpec, 0)
        cubeCoords = np.fliplr(np.array(np.round(cubeCoords[:, :-1]), np.int))

    elif mode in ['pix', 'pixel']:
        assert xyorig, 'if mode==pix, xyorig must be defined.'
        x = coords[:, 0]
        y = coords[:, 1]

        assert shape, 'if mode==pix, shape must be defined.'
        shape = np.atleast_1d(shape)

        if xyorig == 'center':
            yMid, xMid = shape / 2.
            xCube = np.round(xMid + x)
            yCube = np.round(yMid + y)
        elif xyorig == 'lower':
            xCube = np.round(x)
            yCube = np.round(y)
        else:
            raise ValueError('xyorig must be center or lower.')

        cubeCoords = np.array([yCube, xCube], np.int).T

    else:
        raise ValueError('mode must be pix or sky.')

    if shape is not None:
        if ((cubeCoords < 0).any() or
                (cubeCoords[:, 0] > (shape[0] - 1)).any() or
                (cubeCoords[:, 1] > (shape[1] - 1)).any()):
            raise MarvinError('some indices are out of limits.'
                              '``xyorig`` is currently set to "{0}". '
                              'Try setting ``xyorig`` to "{1}".'
                              .format(xyorig, 'center' if xyorig is 'lower' else 'lower'))

    return cubeCoords


def mangaid2plateifu(mangaid, mode='auto', drpall=None, drpver=None):
    """Returns the plate-ifu for a certain mangaid.

    Uses either the DB or the drpall file to determine the plate-ifu for
    a mangaid. If more than one plate-ifu are available for a certain ifu,
    and ``mode='drpall'``, the one with the higher SN2 (calculated as the sum
    of redSN2 and blueSN2) will be used. If ``mode='db'``, the most recent one
    will be used.

    Parameters:
        mangaid (str):
            The mangaid for which the plate-ifu will be returned.
        mode ({'auto', 'drpall', 'db', 'remote'}):
            If `'drpall'` or ``'db'``, the  drpall file or the local database,
            respectively, will be used. If ``'remote'``, a request to the API
            will be issued. If ``'auto'``, the local modes will be tried before
            the remote mode.
        drpall (str or None):
            The path to the drpall file to use. If None, the file in
            ``config.drpall`` will be used.
        drpver (str or None):
            The DRP version to use. If None, the one in ``config.drpver`` will
            be used. If ``drpall`` is defined, this value is ignored.

    Returns:
        plateifu (str):
            The plate-ifu string for the input ``mangaid``.

    """

    from marvin import config, marvindb
    from marvin.api.api import Interaction

    # The modes and order over which the auto mode will loop.
    autoModes = ['db', 'drpall', 'remote']

    assert mode in autoModes + ['auto'], 'mode={0} is not valid'.format(mode)

    config_drpver, __ = config.lookUpVersions()
    drpver = drpver if drpver else config_drpver
    drpall = drpall if drpall else config._getDrpAllPath(drpver=drpver)

    if mode == 'drpall':

        if not drpall:
            raise ValueError('no drpall file can be found.')

        # Loads the drpall table if it was not cached from a previos session.
        if drpver not in drpTable:
            drpTable[drpver] = table.Table.read(drpall)

        mangaids = np.array([mm.strip() for mm in drpTable[drpver]['mangaid']])

        plateifus = drpTable[drpver][np.where(mangaids == mangaid)]

        if len(plateifus) > 1:
            warnings.warn('more than one plate-ifu found for mangaid={0}. '
                          'Using the one with the highest SN2.'.format(mangaid),
                          MarvinUserWarning)
            plateifus = plateifus[
                [np.argmax(plateifus['bluesn2'] + plateifus['redsn2'])]]

        if len(plateifus) == 0:
            raise ValueError('no plate-ifus found for mangaid={0}'.format(mangaid))

        return plateifus['plateifu'][0]

    elif mode == 'db':

        if not marvindb.isdbconnected:
            raise MarvinError('no DB connection found')

        if not drpver:
            raise MarvinError('drpver not set.')

        cubes = marvindb.session.query(marvindb.datadb.Cube).join(
            marvindb.datadb.PipelineInfo, marvindb.datadb.PipelineVersion).filter(
                marvindb.datadb.Cube.mangaid == mangaid,
                marvindb.datadb.PipelineVersion.version == drpver).all()

        if len(cubes) == 0:
            raise ValueError('no plate-ifus found for mangaid={0}'.format(mangaid))
        elif len(cubes) > 1:
            warnings.warn('more than one plate-ifu found for mangaid={0}. '
                          'Using a the one with the higest SN2'.format(mangaid),
                          MarvinUserWarning)
            total_sn2 = [float(cube.header['BLUESN2']) + float(cube.header['REDSN2'])
                         for cube in cubes]
            cube = cubes[np.argmax(total_sn2)]
        else:
            cube = cubes[0]

        return '{0}-{1}'.format(cube.plate, cube.ifu.name)

    elif mode == 'remote':

        try:
            # response = Interaction('api/general/mangaid2plateifu/{0}/'.format(mangaid))
            url = marvin.config.urlmap['api']['mangaid2plateifu']['url']
            response = Interaction(url.format(mangaid=mangaid))
        except MarvinError as e:
            raise MarvinError('API call to mangaid2plateifu failed: {0}'.format(e))
        else:
            plateifu = response.getData(astype=str)

        if not plateifu:
            if 'error' in response.results and response.results['error']:
                raise MarvinError(response.results['error'])
            else:
                raise MarvinError('API call to mangaid2plateifu failed with error unknown.')

        return plateifu

    elif mode == 'auto':

        for mm in autoModes:
            try:
                plateifu = mangaid2plateifu(mangaid, mode=mm, drpver=drpver, drpall=drpall)
                return plateifu
            except:
                continue

        raise MarvinError(
            'mangaid2plateifu was not able to find a plate-ifu for '
            'mangaid={0} either local or remotely.'.format(mangaid))


def findClosestVector(point, arr_shape=None, pixel_shape=None, xyorig=None):
    '''
    Finds the closest vector of array coordinates (x, y) from an input vector of pixel coordinates (x, y).

    Parameters:
        point : tuple
            Original point of interest in pixel units, order of (x,y)
        arr_shape : tuple
            Shape of data array in (x,y) order
        pixel_shape : tuple
            Shape of image in pixels in (x,y) order
        xyorig : str
            Indicates the origin point of coordinates.  Set to "relative" switches to an array coordinate
            system relative to galaxy center.  Default is absolute array coordinates (x=0, y=0) = upper left corner

    Returns:
        minind : tuple
            A tuple of array coordinates in x, y order
    '''

    # set as numpy arrays
    arr_shape = np.array(arr_shape, dtype=int)
    pixel_shape = np.array(pixel_shape, dtype=int)

    # compute midpoints
    xmid, ymid = arr_shape/2
    xpixmid, ypixmid = pixel_shape/2

    # default absolute array coordinates
    xcoords = np.array([0, arr_shape[0]], dtype=int)
    ycoords = np.array([0, arr_shape[1]], dtype=int)

    # split x,y coords and pixel coords
    x1, x2 = xcoords
    y1, y2 = ycoords
    xpix, ypix = pixel_shape

    # build interpolates between array coordinates and pixel coordinates
    points = [[x1, y1], [x1, y2], [xmid, ymid], [x2, y1], [x2, y2]]
    values = [[0, ypix], [0, 0], [xpixmid, ypixmid], [xpix, ypix], [xpix, 0]]  # full image
    # values = [[xpixmid-xmid, ypixmid+ymid], [xpixmid-xmid, ypixmid-ymid], [xpixmid, ypixmid], [xpixmid+xmid, ypixmid+ymid], [xpixmid+xmid, ypixmid-ymid]]  # pixels based on arr_shape
    #values = [[xpixmid-x2, ypixmid+y2], [xpixmid-x2, ypixmid-y2], [xpixmid, ypixmid], [xpixmid+x2, ypixmid+y2], [xpixmid+x2, ypixmid-y2]]  # pixels based on arr_shape

    # make 2d array of array indices in absolute or (our) relative coordindates
    arrinds = np.mgrid[x1:x2, y1:y2].swapaxes(0, 2).swapaxes(0, 1)
    # interpolate a new 2d pixel coordinate array
    final = griddata(points, values, arrinds)

    # find minimum array vector closest to input coordinate point
    diff = np.abs(point - final)
    prod = diff[:, :, 0]*diff[:, :, 1]
    minind = np.unravel_index(prod.argmin(), arr_shape)

    # toggle relative array coordinates
    if xyorig in ['relative', 'center']:
        minind = np.array(minind, dtype=int)
        xmin = minind[0] - xmid
        ymin = ymid - minind[1]
        minind = (xmin, ymin)

    return minind


def getWCSFromPng(image):
    ''' Extracts any WCS info from the metadata of a PNG image

    Extracts the WCS metadata info from the PNG optical
    image of the galaxy using PIL (Python Imaging Library).
    Converts it to an Astropy WCS object.

    Parameters:
        image (str):
            The full path to the image

    Returns:
        pngwcs (WCS):
            an Astropy WCS object

    '''

    pngwcs = None
    try:
        image = PIL.Image.open(image)
    except Exception as e:
        raise MarvinError('Cannot open image {0}: {1}'.format(image, e))

    # get metadata
    meta = image.info if image else None

    # parse the image metadata
    mywcs = {}
    if meta and 'WCSAXES' in meta.keys():
        for key, val in meta.items():
            try:
                val = float(val)
            except Exception as e:
                pass
            mywcs.update({key: val})

        tmp = mywcs.pop('WCSAXES')

    # Construct Astropy WCS
    if mywcs:
        pngwcs = wcs.WCS(mywcs)

    # Close the image
    image.close()

    return pngwcs


def convertImgCoords(coords, image, to_pix=None, to_radec=None):
    ''' Transform the WCS info in an image

    Convert image pixel coordinates to RA/Dec based on
    PNG image metadata or vice_versa

    Parameters:
        coords (tuple):
            The input coordindates to transform
        image (str):
            The full path to the image
        to_pix (bool):
            Set to convert to pixel coordinates
        to_radec (bool):
            Set to convert to RA/Dec coordinates

    Returns:
        newcoords (tuple):
            Tuple of either (x, y) pixel coordinates
            or (RA, Dec) coordinates

    '''

    try:
        wcs = getWCSFromPng(image)
    except Exception as e:
        raise MarvinError('Cannot get wcs info from image {0}: {1}'.format(image, e))

    if to_radec:
        try:
            newcoords = wcs.all_pix2world([coords], 1)[0]
        except AttributeError as e:
            raise MarvinError('Cannot convert coords to RA/Dec.  No wcs! {0}'.format(e))
    if to_pix:
        try:
            newcoords = wcs.all_world2pix([coords], 1)[0]
        except AttributeError as e:
            raise MarvinError('Cannot convert coords to image pixels.  No wcs! {0}'.format(e))
    return newcoords


def parseIdentifier(galid):
    ''' Determines if a string input is a plate, plateifu, or manga-id

    Parses a string object id and determines whether it is a
    plate ID, a plate-IFU, or MaNGA-ID designation.

    Parameters:
        galid (str):
            The string of an id name to parse

    Returns:
        idtype (str):
            String indicating either plate, plateifu, mangaid, or None

    '''

    galid = str(galid)
    hasdash = '-' in galid

    if hasdash:
        galidsplit = galid.split('-')

        if int(galidsplit[0]) > 6500:
            idtype = 'plateifu'
        else:
            idtype = 'mangaid'
    else:
        # check for plate
        if galid.isdigit():
            if len(galid) > 3:
                idtype = 'plate'
            else:
                idtype = None
        else:
            idtype = None

    return idtype


def getSpaxelXY(cube, plateifu, x, y):
    """Gets and spaxel from a cube in the DB.

    This function is mostly intended for internal use.

    Parameters:
        cube (SQLAlchemy object):
            The SQLAlchemy object representing the cube from which to extract
            the spaxel.
        plateifu (str):
            The corresponding plateifu of ``cube``.
        x,y (int):
            The coordinates of the spaxel in the database.

    Returns:
        spaxel (SQLAlchemy object):
            The SQLAlchemy spaxel with coordinates ``(x, y)``.

    """

    import sqlalchemy

    mdb = marvin.marvindb

    try:
        spaxel = mdb.session.query(mdb.datadb.Spaxel).filter_by(cube=cube, x=x, y=y).one()
    except sqlalchemy.orm.exc.NoResultFound as e:
        raise MarvinError('Could not retrieve spaxel for plate-ifu {0} at position {1},{2}: No Results Found: {3}'.format(plateifu, x, y, e))
    except Exception as e:
        raise MarvinError('Could not retrieve cube for plate-ifu {0} at position {1},{2}: Unknown exception: {3}'.format(plateifu, x, y, e))

    return spaxel


def getDapRedux(release=None):
    ''' Retrieve SAS url link to the DAP redux directory

    Parameters:
        release (str):
            The release version of the data to download.
            Defaults to Marvin config.release.

    Returns:
        dapredux (str):
            The full redux path to the DAP MAPS
    '''

    if not Path:
        raise MarvinError('sdss_access is not installed')
    else:
        sdss_path = Path()

    release = release or marvin.config.release
    drpver, dapver = marvin.config.lookUpVersions(release=release)
    # hack a url version of MANGA_SPECTRO_ANALYSIS
    dapdefault = sdss_path.dir('mangadefault', drpver=drpver, dapver=dapver, plate=None, ifu=None)
    dappath = dapdefault.rsplit('/', 2)[0]
    dapredux = sdss_path.url('', full=dappath)
    return dapredux


def getDefaultMapPath(**kwargs):
    ''' Retrieve the default Maps path

    Uses sdss_access Path to generate a url download link to the
    default MAPS file for a given MPL.

    Parameters:
        release (str):
            The release version of the data to download.
            Defaults to Marvin config.release.
        plate (int):
            The plate id
        ifu (int):
            The ifu number
        bintype (str):
            The bintype of the default file to grab. Defaults to MAPS
        daptype (str):
            The daptype of the default map to grab.  Defaults to SPX-MILESHC

    Returns:
        maplink (str):
            The sas url to download the default maps file
    '''

    if not Path:
        raise MarvinError('sdss_access is not installed')
    else:
        sdss_path = Path()

    # Get kwargs
    release = kwargs.get('release', marvin.config.release)
    drpver, dapver = marvin.config.lookUpVersions(release=release)
    plate = kwargs.get('plate', None)
    ifu = kwargs.get('ifu', None)
    daptype = kwargs.get('daptype', 'SPX-GAU-MILESHC')
    bintype = kwargs.get('bintype', 'MAPS')

    # get the sdss_path name by MPL
    # TODO: this is likely to break in future MPL/DRs. Just a heads up.
    if '4' in release:
        name = 'mangadefault'
    elif '5' in release:
        name = 'mangadap5'
    else:
        return None

    # construct the url link to default maps file
    maplink = sdss_path.url(name, drpver=drpver, dapver=dapver, mpl=release,
                            plate=plate, ifu=ifu, daptype=daptype, mode=bintype)
    return maplink


def downloadList(inputlist, dltype='cube', **kwargs):
    ''' Download a list of MaNGA objects

    Uses sdss_access to download a list of objects
    via rsync.  Places them in your local sas path mimicing
    the Utah SAS.

    i.e. $SAS_BASE_DIR/mangawork/manga/spectro/redux

    Can download cubes, rss files, maps, mastar cubes, png images, default maps, or
    the entire plate directory.

    Parameters:
        inputlist (list):
            Required.  A list of objects to download.  Must be a list of plate IDs,
            plate-IFUs, or manga-ids
        dltype ({'cube', 'map', 'image', 'rss', 'mastar', 'default', 'plate'}):
            Indicated type of object to download.  Can be any of
            plate, cube, imagea, mastar, rss, map, or default (default map).
            If not specified, the dltype defaults to cube.
        release (str):
            The MPL/DR version of the data to download.
            Defaults to Marvin config.release.
        bintype (str):
            The bin type of the DAP maps to download. Defaults to *
        binmode (str):
            The bin mode of the DAP maps to download. Defaults to *
        n (int):
            The plan id number [1-12] of the DAP maps to download. Defaults to *
        daptype (str):
            The daptype of the default map to grab.  Defaults to *
        dir3d (str):
            The directory where the images are located.  Either 'stack' or 'mastar'. Defaults to *
        verbose (bool):
            Turns on verbosity during rsync
        limit (int):
            A limit to the number of items to download
    Returns:
        NA: Downloads

    '''

    # Get some possible keywords
    # Necessary rsync variables:
    #   drpver, plate, ifu, dir3d, [mpl, dapver, bintype, n, mode]
    verbose = kwargs.get('verbose', None)
    as_url = kwargs.get('as_url', None)
    release = kwargs.get('release', marvin.config.release)
    drpver, dapver = marvin.config.lookUpVersions(release=release)
    bintype = kwargs.get('bintype', '*')
    binmode = kwargs.get('binmode', '*')
    daptype = kwargs.get('daptype', '*')
    dir3d = kwargs.get('dir3d', '*')
    n = kwargs.get('n', '*')
    limit = kwargs.get('limit', None)

    # check for sdss_access
    if not RsyncAccess:
        raise MarvinError('sdss_access not installed.')

    # Assert correct dltype
    dltype = 'cube' if not dltype else dltype
    assert dltype in ['plate', 'cube', 'mastar', 'rss', 'map', 'image',
                      'default'], 'dltype must be one of plate, cube, mastar, image, rss, map, default'

    # Assert correct dir3d
    if dir3d != '*':
        assert dir3d in ['stack', 'mastar'], 'dir3d must be either stack or mastar'

    # Parse and retrieve the input type and the download type
    idtype = parseIdentifier(inputlist[0])
    if not idtype:
        raise MarvinError('Input list must be a list of plates, plate-ifus, or mangaids')

    # Set download type
    if dltype == 'cube':
        name = 'mangacube'
    elif dltype == 'rss':
        name = 'mangarss'
    elif dltype == 'default':
        name = 'mangadefault'
    elif dltype == 'plate':
        name = 'mangaplate'
    elif dltype == 'map':
        if '4' in release:
            name = 'mangamap'
        elif '5' in release:
            name = 'mangadap5'
    elif dltype == 'mastar':
        name = 'mangamastar'
    elif dltype == 'image':
        name = 'mangaimage'

    # create rsync
    rsync_access = RsyncAccess(label='marvin_download', verbose=verbose)
    rsync_access.remote()

    # Add objects
    for item in inputlist:
        if idtype == 'mangaid':
            try:
                plateifu = mangaid2plateifu(item)
            except MarvinError as e:
                plateifu = None
            else:
                plateid, ifu = plateifu.split('-')
        elif idtype == 'plateifu':
            plateid, ifu = item.split('-')
        elif idtype == 'plate':
            plateid = item
            ifu = '*'

        rsync_access.add(name, plate=plateid, drpver=drpver, ifu=ifu, dapver=dapver, dir3d=dir3d,
                         mpl=release, bintype=bintype, n=n, mode=binmode, daptype=daptype)

    # set the stream
    try:
        rsync_access.set_stream()
    except AccessError as e:
        raise MarvinError('Error with sdss_access rsync.set_stream. AccessError: {0}'.format(e))

    # get the list and download
    listofitems = rsync_access.get_urls() if as_url else rsync_access.get_paths()
    rsync_access.commit(limit=limit)


def get_drpall_row(plateifu, drpver=None, drpall=None):
    """Returns a dictionary from drpall matching the plateifu."""

    from marvin import config

    if drpall:
        drpall_table = table.Table.read(drpall)
    else:
        config_drpver, __ = config.lookUpVersions()
        drpver = drpver if drpver else config_drpver
        if drpver not in drpTable:
            drpall = drpall if drpall else config._getDrpAllPath(drpver=drpver)
            drpTable[drpver] = table.Table.read(drpall)
        drpall_table = drpTable[drpver]

    row = drpall_table[drpall_table['plateifu'] == plateifu]

    if len(row) != 1:
        raise ValueError('{0} results found for {1} in drpall table'.format(len(row), plateifu))

    return row[0]


def _db_row_to_dict(row, remove_columns=False):
    """Converts a DB object to a dictionary."""

    from sqlalchemy.inspection import inspect as sa_inspect
    from sqlalchemy.ext.hybrid import hybrid_property
    from sqlalchemy.orm.attributes import InstrumentedAttribute

    row_dict = collections.OrderedDict()

    columns = row.__table__.columns.keys()

    mapper = sa_inspect(row.__class__)
    for key, item in mapper.all_orm_descriptors.items():

        if isinstance(item, hybrid_property):
            columns.append(key)

    for col in columns:
        if remove_columns and col in remove_columns:
            continue
        row_dict[col] = getattr(row, col)

    return row_dict


def get_nsa_data(mangaid, source='nsa', mode='auto', drpver=None, drpall=None):
    """Returns a dictionary of NSA data from the DB or from the drpall file.

    Parameters:
        mangaid (str):
            The mangaid of the target for which the NSA information will be returned.
        source ({'nsa', 'drpall'}):
            The data source. If ``source='nsa'``, the full NSA catalogue from the DB will
            be used. If ``source='drpall'``, the subset of NSA columns included in the drpall
            file will be returned.
        mode ({'auto', 'local', 'remote'}):
            See :ref:`mode-decision-tree`.
        drpver (str or None):
            The version of the DRP to use, if ``source='drpall'``. If ``None``, uses the
            version set by ``marvin.config.release``.
        drpall (str or None):
            A path to the drpall file to use if ``source='drpall'``. If not defined, the
            default drpall file matching ``drpver`` will be used.

    Returns:
        nsa_data (dict):
            A dictionary containing the columns and values from the NSA catalogue for
            ``mangaid``.

    """

    from marvin import config, marvindb
    from marvin.core.core import DotableCaseInsensitive

    valid_modes = ['auto', 'local', 'remote']
    assert mode in valid_modes, 'mode must be one of {0}'.format(valid_modes)

    valid_sources = ['nsa', 'drpall']
    assert source in valid_sources, 'source must be one of {0}'.format(valid_sources)

    log.debug('get_nsa_data: getting NSA data for mangaid=%r with source=%r, mode=%r',
              mangaid, source, mode)

    if mode == 'auto':
        log.debug('get_nsa_data: running auto mode mode.')
        try:
            nsa_data = get_nsa_data(mangaid, mode='local', source=source,
                                    drpver=drpver, drpall=drpall)
            return nsa_data
        except MarvinError as ee:
            log.debug('get_nsa_data: local mode failed with error %s', str(ee))
            try:
                nsa_data = get_nsa_data(mangaid, mode='remote', source=source,
                                        drpver=drpver, drpall=drpall)
                return nsa_data
            except MarvinError as ee:
                raise MarvinError('get_nsa_data: failed to get NSA data for mangaid=%r in '
                                  'auto mode with with error: %s', mangaid, str(ee))

    elif mode == 'local':

        if source == 'nsa':

            if config.db is not None:

                session = marvindb.session
                sampledb = marvindb.sampledb

                nsa_row = session.query(sampledb.NSA).join(sampledb.MangaTargetToNSA,
                                                           sampledb.MangaTarget).filter(
                    sampledb.MangaTarget.mangaid == mangaid).all()

                if len(nsa_row) == 1:
                    return DotableCaseInsensitive(
                        _db_row_to_dict(nsa_row[0], remove_columns=['pk', 'catalogue_pk']))
                elif len(nsa_row) > 1:
                    warnings.warn('get_nsa_data: multiple NSA rows found for mangaid={0}. '
                                  'Using the first one.'.format(mangaid), MarvinUserWarning)
                    return DotableCaseInsensitive(
                        _db_row_to_dict(nsa_row[0], remove_columns=['pk', 'catalogue_pk']))
                elif len(nsa_row) == 0:
                    raise MarvinError('get_nsa_data: cannot find NSA row for mangaid={0}'
                                      .format(mangaid))

            else:

                raise MarvinError('get_nsa_data: cannot find a valid DB connection.')

        elif source == 'drpall':

            plateifu = mangaid2plateifu(mangaid, drpver=drpver, drpall=drpall, mode='drpall')
            log.debug('get_nsa_data: found plateifu=%r for mangaid=%r', plateifu, mangaid)

            drpall_row = get_drpall_row(plateifu, drpall=drpall, drpver=drpver)

            nsa_data = collections.OrderedDict()
            for col in drpall_row.colnames:
                if col.startswith('nsa_'):
                    value = drpall_row[col]
                    if isinstance(value, np.ndarray):
                        value = value.tolist()
                    else:
                        value = np.asscalar(value)
                    nsa_data[col[4:]] = value

            return DotableCaseInsensitive(nsa_data)

    elif mode == 'remote':

        from marvin.api.api import Interaction

        try:
            if source == 'nsa':
                request_name = 'nsa_full'
            else:
                request_name = 'nsa_drpall'
            url = marvin.config.urlmap['api'][request_name]['url']
            response = Interaction(url.format(mangaid=mangaid))
        except MarvinError as ee:
            raise MarvinError('API call to {0} failed: {1}'.format(request_name, str(ee)))
        else:
            if response.results['status'] == 1:
                return DotableCaseInsensitive(collections.OrderedDict(response.getData()))
            else:
                raise MarvinError('get_nsa_data: %s', response['error'])


def _check_file_parameters(obj1, obj2):
    for param in ['plateifu', 'mangaid', 'plate', '_release', 'drpver', 'dapver']:
        assert_msg = ('{0} is different between {1} {2}:\n {1}.{0}: {3} {2}.{0}:{4}'
                      .format(param, obj1.__repr__, obj2.__repr__, getattr(obj1, param),
                              getattr(obj2, param)))
        assert getattr(obj1, param) == getattr(obj2, param), assert_msg


def get_plot_params(dapver, prop):
    """Returns default plotting parameters for a property."""
    params = get_default_plot_params(dapver)

    if 'vel' in prop:
        key = 'vel'
    elif 'sigma' in prop:
        key = 'sigma'
    else:
        key = 'default'

    return params[key]




# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-06-20 16:36:37
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-07-30 18:19:10

from __future__ import print_function, division, absolute_import
from marvin.utils.general.images import getImagesByList, getImagesByPlate, getRandomImages, getDir3d, showImage
from marvin.tests.conftest import Galaxy, tempafile
from marvin.tests import marvin_test_if
from marvin.core.exceptions import MarvinError, MarvinUserWarning
import pytest
import os
import warnings


try:
    from sdss_access import RsyncAccess, AccessError
except ImportError:
    Path = None
    RsyncAccess = None


imagelist = ['8485-1901', '7443-12701', '7443-1901']
newgals = ['7495-1901']


@pytest.fixture(scope='function')
def rsync(mode):
    ''' fixture to create generic rsync object '''

    rsync = RsyncAccess(label='marvin_getlist', verbose=False)
    if mode != 'local':
        rsync.remote()
    yield rsync
    rsync.reset()
    rsync = None


localredux = os.getenv('MANGA_SPECTRO_REDUX')
remoteredux = 'https://sdss@dtn01.sdss.org/sas/mangawork/manga/spectro/redux'
remoteurl = 'https://data.sdss.org/sas/mangawork/manga/spectro/redux'
bases = [localredux, remoteredux, remoteurl]
rmodes = ['full', 'url']


@pytest.fixture()
def base(mode, asurl):
    if asurl is False:
        return localredux
    else:
        if mode != 'local':
            return remoteredux
        else:
            return remoteurl


@pytest.fixture(scope='session', params=newgals)
def newgalaxy(request, maindb, get_params, set_sasurl):
    release, bintype, template = get_params

    gal = Galaxy(request.param)
    gal.set_params(bintype=bintype, template=template)
    gal.set_filepaths()
    yield gal


@pytest.fixture()
def get_cube(newgalaxy, rsync):
    if not os.path.isfile(newgalaxy.cubepath):
        rsync.add('mangacube', **newgalaxy.access_kwargs)
        rsync.set_stream()
        rsync.commit()
    yield newgalaxy


@pytest.fixture(params=rmodes)
def asurl(request):
    if request.param == 'full':
        return False
    elif request.param == 'url':
        return True


@pytest.fixture()
def make_paths(request, rsync, mode, asurl, release):
    inputs = request.param if hasattr(request, 'param') else None
    rmode = 'url' if asurl else 'full'
    fullpaths = []
    inputs = inputs if inputs else imagelist
    for plateifu in inputs:
        gal = Galaxy(plateifu)
        gal.set_params(release=release)
        gal.set_filepaths()
        if mode == 'local':
            path = rsync.__getattribute__(rmode)('mangaimage', **gal.access_kwargs)
            fullpaths.append(path)
        else:
            rsync.add('mangaimage', **gal.access_kwargs)
            rsync.set_stream()
            path = rsync.get_urls() if asurl else rsync.get_paths()
            fullpaths.extend(path)
    return fullpaths


class TestImagesGetDir3d(object):

    @pytest.mark.parametrize('expval', [('stack')])
    def test_getdir3d(self, galaxy, release, expval, mode, db):
        dir3d = getDir3d(galaxy.plateifu, mode=mode, release=release)
        assert expval == dir3d

    @pytest.mark.parametrize('expval', [('stack')])
    def test_getdir3d_plate(self, galaxy, release, expval, mode, db):
        dir3d = getDir3d(galaxy.plate, mode=mode, release=release)
        assert expval == dir3d


@pytest.mark.xfail()
@pytest.mark.timeout(40)
class TestImagesByList(object):

    @pytest.mark.parametrize('imglist, mode, errmsg',
                             [('7495-1901', 'local', 'Input must be of type list or Numpy array'),
                              (['nogoodid'], 'local', 'Input must be of type plate-ifu or mangaid'),
                              (imagelist, 'notvalidmode', 'Mode must be either auto, local, or remote')],
                             ids=['notlist', 'badid', 'badmode'])
    def test_failures(self, imglist, mode, errmsg, release):
        with pytest.raises(AssertionError) as cm:
            image = getImagesByList(imglist, mode=mode, release=release)
        assert cm.type == AssertionError
        assert errmsg in str(cm.value)

    def test_get_imagelist(self, make_paths, mode, asurl, release):
        images = getImagesByList(imagelist, mode=mode, as_url=asurl, release=release)
        assert set(make_paths) == set(images)

    # @pytest.mark.parametrize('make_paths', [(['7495-1901'])], indirect=True, ids=['newplateifu'])
    # def test_download(self, monkeymanga, temp_scratch, get_cube):
    #     imgpath = tempafile(get_cube.imgpath, temp_scratch)
    #     #assert os.path.isfile(get_cube.imgpath) is False
    #     assert imgpath.check(file=0) is True
    #     image = getImagesByList([get_cube.plateifu], mode='remote', as_url=True, download=True, release=get_cube.release)
    #     #assert os.path.isfile(get_cube.imgpath) is True
    #     assert imgpath.check(file=1) is True
    #     assert image is None

    # @pytest.mark.parametrize('make_paths', [(['7495-1901'])], indirect=True, ids=['newplateifu'])
    # def test_download_fails(self, monkeymanga, temp_scratch, get_cube):
    #     imgpath = tempafile(get_cube.imgpath, temp_scratch)
    #     assert imgpath.check(file=0) is True
    #     errmsg = 'Download not available when in local mode'
    #     with warnings.catch_warnings(record=True) as cm:
    #         warnings.simplefilter('always')
    #         image = getImagesByList([get_cube.plateifu], mode='local', as_url=True, download=True)
    #     assert cm[-1].category is MarvinUserWarning
    #     assert errmsg in str(cm[-1].message)


class TestImagesByPlate(object):

    @pytest.mark.parametrize('plateid, mode, errmsg',
                             [('8485abcd', 'local', 'Plateid must be a numeric integer value'),
                              (None, 'notvalidmode', 'Mode must be either auto, local, or remote')],
                             ids=['badid', 'badmode'])
    def test_failures(self, galaxy, plateid, mode, errmsg, release):
        plateid = plateid if plateid else galaxy.plate
        with pytest.raises(AssertionError) as cm:
            image = getImagesByPlate(plateid, mode=mode, release=release)
        assert cm.type == AssertionError
        assert errmsg in str(cm.value)

    @pytest.mark.parametrize('make_paths, plate', [(['8485-1901'], '8485')], indirect=['make_paths'], ids=['plateifu'])
    def test_get_imageplate(self, make_paths, plate, mode, asurl, release):
        images = getImagesByPlate(plate, mode=mode, as_url=asurl, release=release)
        assert make_paths[0] in images

    # @pytest.mark.parametrize('make_paths', [(['7495-1901'])], indirect=True, ids=['newplateifu'])
    # def test_download(self, monkeymanga, temp_scratch, get_cube):
    #     imgpath = tempafile(get_cube.imgpath, temp_scratch)
    #     assert imgpath.check(file=0) is True
    #     image = getImagesByPlate(get_cube.plate, mode='remote', as_url=True, download=True)
    #     assert imgpath.check(file=1) is True
    #     assert image is None

    # def test_get_images_download_local_fail(self, monkeymanga, temp_scratch, get_cube):
    #     imgpath = tempafile(get_cube.imgpath, temp_scratch)
    #     assert imgpath.check(file=0) is True
    #     errmsg = 'Download not available when in local mode'
    #     with warnings.catch_warnings(record=True) as cm:
    #         warnings.simplefilter("always")
    #         image = getImagesByPlate(self.new_plate, mode='local', as_url=True, download=True)
    #     self.assertIs(cm[-1].category, MarvinUserWarning)
    #     self.assertIn(errmsg, str(cm[-1].message))


class TestRandomImages(object):

    @pytest.mark.parametrize('mode, errmsg',
                             [('notvalidmode', 'Mode must be either auto, local, or remote')],
                             ids=['badmode'])
    def test_failures(self, mode, errmsg, release):
        with pytest.raises(AssertionError) as cm:
            image = getRandomImages(mode=mode, release=release)
        assert cm.type == AssertionError
        assert errmsg in str(cm.value)

    @pytest.mark.parametrize('num', [(10), (5)], ids=['num10', 'num5'])
    def test_get_image_random(self, base, num, mode, asurl, release):
        images = getRandomImages(num=num, mode=mode, as_url=asurl, release=release)
        assert images is not None
        assert num == len(images)
        assert isinstance(images, list) is True
        assert base in images[0]


class TestShowImage(object):

    def _assert_image(self, galaxy, image):
        assert image is not None
        assert image.size == (562, 562)
        assert image.format == 'PNG'
        assert str(galaxy.plate) in image.filename
        assert galaxy.ifu in image.filename

    @pytest.mark.parametrize('return_image', [(True), (False)], ids=['returnyes', 'returnno'])
    def test_show_image(self, galaxy, mode, return_image):
        image = showImage(plateifu=galaxy.plateifu, mode=mode, release=galaxy.release, return_image=return_image, show_image=False)
        if return_image:
            self._assert_image(galaxy, image)
        else:
            assert image is None

        return image

    @pytest.mark.parametrize('param, error, errmsg',
                             [({'mode': 'notvalidmode'}, AssertionError, 'Mode must be either auto, local, or remote'),
                              ({}, AssertionError, 'A filepath or plateifu must be specified!'),
                              ({'plateifu': '8485-1905'}, MarvinError, 'Error: remote filepath'),
                              ({'path': '/tmp/image.png'}, MarvinError, 'Error: local filepath /tmp/image.png does not exist.'),
                              ({'path': ['/tmp/image.png', '/tmp/image1.png']}, MarvinError, 'showImage currently only works on a single input at a time')],
                             ids=['badmode', 'noinput', 'badplateifu', 'badfile', 'toomany'])
    def test_failures(self, param, error, errmsg, release):
        if 'mode' not in param:
            param.update({'mode': 'auto'})

        with pytest.raises(error) as cm:
            image = showImage(release=release, **param)
        assert cm.type == error
        assert errmsg in str(cm.value)

    def test_withpaths(self, galaxy, mode):
        if mode != 'local':
            galaxy.set_params()
            galaxy.set_filepaths(pathtype='url')
        image = showImage(path=galaxy.imgpath, mode=mode, return_image=True, show_image=False)
        self._assert_image(galaxy, image)

    @pytest.mark.parametrize('base, mode, errmsg',
                             [(localredux, 'remote', 'Local path not allowed in remote mode'),
                              (remoteurl, 'local', 'Remote url path not allowed in local mode')],
                             ids=['remoteuserdir', 'localhttp'])
    def test_path_fails_wrongmodes(self, base, galaxy, mode, errmsg):
        path = os.path.join(base, galaxy.get_location(galaxy.imgpath))
        with pytest.raises(MarvinError) as cm:
            image = showImage(release=galaxy.release, path=path, mode=mode)
        assert cm.type == MarvinError
        assert errmsg in str(cm.value)



from __future__ import division, print_function

import os
import re
from glob import glob
from os.path import join
from random import choice, sample
from re import compile

try:
    from ConfigParser import RawConfigParser
except ImportError:
    from configparser import RawConfigParser

"""
Module for constructing paths to SDSS files.

Example use case:

    from sdss_access.path import Path
    sdss_path = Path()
    filename = sdss_path.full('photoObj', run=94, rerun='301', camcol=1, field=100)

Depends on the tree product. In particular requires path templates in:
  $TREE_DIR/data/sdss_paths.ini
"""


class BasePath(object):
    """Class for construction of paths in general.

    Attributes
    ----------
    templates : dict
        The set of templates read from the configuration file.
    """

    _netloc = {"dtn":"sdss@dtn01.sdss.org", "sdss":"data.sdss.org", "mirror":"data.mirror.sdss.org"}

    def __init__(self, pathfile, mirror=False, public=False, verbose=False):
        self.mirror = mirror
        self.public = public
        self.verbose = verbose
        self.set_netloc()
        self.set_remote_base()
        self._pathfile = pathfile
        self._config = RawConfigParser()
        self._config.optionxform = str
        self.templates = dict()
        self._input_templates()
        return

    def _input_templates(self):
        """Read the path template file.
        """
        foo = self._config.read([self._pathfile])
        if len(foo) == 1:
            for k, v in self._config.items('paths'):
                self.templates[k] = v
        else:
            raise ValueError("Could not read {0}!".format(self._pathfile))
        return

    def dir(self, filetype, **kwargs):
        """Return the directory containing a file of a given type.

        Parameters
        ----------
        filetype : str
            File type parameter.

        Returns
        -------
        dir : str
            Directory containing the file.
        """
        full = kwargs.get('full', self.full(filetype, **kwargs))
        return os.path.dirname(full)

    def name(self, filetype, **kwargs):
        """Return the directory containing a file of a given type.

        Parameters
        ----------
        filetype : str
            File type parameter.

        Returns
        -------
        name : str
            Name of a file with no directory information.
        """
        full = kwargs.get('full', self.full(filetype, **kwargs))
        return os.path.basename(full)

    def exists(self, filetype, **kwargs):
        '''Checks if the given type of file exists

        Parameters
        ----------
        filetype : str
            File type parameter.

        Returns
        -------
        exists : bool
            Boolean indicating if the file exists on disk.

        '''
        full = kwargs.get('full', self.full(filetype, **kwargs))
        return os.path.isfile(full)

    def expand(self, filetype, **kwargs):
        ''' Expand a wildcard path locally

        Parameters
        ----------
        filetype : str
            File type parameter.

        as_url: bool
            Boolean to return SAS urls

        refine: str
            Regular expression string to filter the list of files by
            before random selection

        Returns
        -------
        expand : list
            List of expanded full paths of the given type.

        '''
        full = kwargs.get('full', self.full(filetype, **kwargs))
        assert '*' in full, 'Wildcard must be present in full path'
        files = glob(full)

        # return as urls?
        as_url = kwargs.get('as_url', None)
        newfiles = [self.url('', full=full) for full in files] if as_url else files

        # optionally refine the results
        refine = kwargs.get('refine', None)
        if refine:
            newfiles = self.refine(newfiles, refine, **kwargs)

        return newfiles

    def any(self, filetype, **kwargs):
        ''' Checks if the local directory contains any of the type of file

        Parameters
        ----------
        filetype : str
            File type parameter.

        Returns
        -------
        any : bool
            Boolean indicating if the any files exist in the expanded path on disk.

        '''
        expanded_files = self.expand(filetype, **kwargs)
        return any(expanded_files)

    def one(self, filetype, **kwargs):
        ''' Returns random one of the given type of file

        Parameters
        ----------
        filetype : str
            File type parameter.

        as_url: bool
            Boolean to return SAS urls

        refine: str
            Regular expression string to filter the list of files by
            before random selection

        Returns
        -------
        one : str
            Random file selected from the expanded list of full paths on disk.

        '''
        expanded_files = self.expand(filetype, **kwargs)
        isany = self.any(filetype, **kwargs)
        return choice(expanded_files) if isany else None

    def random(self, filetype, **kwargs):
        ''' Returns random number of the given type of file

        Parameters
        ----------
        filetype : str
            File type parameter.

        num : int
            The number of files to return

        as_url: bool
            Boolean to return SAS urls

        refine: str
            Regular expression string to filter the list of files by
            before random selection

        Returns
        -------
        random : list
            Random file selected from the expanded list of full paths on disk.

        '''
        expanded_files = self.expand(filetype, **kwargs)
        isany = self.any(filetype, **kwargs)
        if isany:
            # get the desired number
            num = kwargs.get('num', 1)
            assert num <= len(expanded_files), 'Requested number must be larger the sample.  Reduce your number.'
            return sample(expanded_files, num)
        else:
            return None

    def refine(self, filelist, regex, filterdir='out', **kwargs):
        ''' Returns a list of files filterd by a regular expression

        Parameters
        ----------
        filelist : list
            A list of files to filter on.

        regex : str
            The regular expression string to filter your list

        filterdir: {'in', 'out'}
            Indicates the filter to be inclusive or exclusive
            'out' removes the items satisfying the regular expression
            'in' keeps the items satisfying the regular expression

        Returns
        -------
        refine : list
            A file list refined by an input regular expression.

        '''
        assert filelist, 'Must provide a list of filenames to refine on'
        assert regex, 'Must provide a regular expression to refine the file list'
        r = compile(regex)

        # icheck filter direction; default is out
        assert filterdir in ['in', 'out'], 'Filter direction must be either "in" or "out"'
        if filterdir == 'out':
            subset = list(filter(lambda i: r.search(i), filelist))
        elif filterdir == 'in':
            subset = list(filter(lambda i: not r.search(i), filelist))
        return subset

    def full(self, filetype, **kwargs):
        """Return the full name of a given type of file.

        Parameters
        ----------
        filetype : str
            File type parameter.

        Returns
        -------
        full : str
            The full path to the file.
        """
        try:
            template = self.templates[filetype]
        except KeyError:
            return None
        # Now replace {} items
        template = template.format(**kwargs)

        # Now replace environmental variables
        envvars = re.findall(r"\$\w+", template)
        for envvar in envvars:
            try:
                value = os.environ[envvar[1:]]
            except KeyError:
                return None
            template = re.sub("\\" + envvar, value, template)

        # Now call special functions as appropriate
        functions = re.findall(r"\%\w+", template)
        for function in functions:
            try:
                method = getattr(self, function[1:])
            except AttributeError:
                return None
            value = method(filetype, **kwargs)
            template = re.sub(function, value, template)

        return template

    def set_netloc(self, netloc=None, sdss=None, dtn=None):
        self.netloc =  netloc if netloc else self._netloc["sdss"] if sdss else self._netloc["dtn"] if dtn  else self._netloc["mirror"] if self.mirror else self._netloc["sdss"]

    def set_remote_base(self, scheme=None):
        self.remote_base = self.get_remote_base(scheme=scheme) if scheme else self.get_remote_base()

    def get_remote_base(self, scheme="https"):
        return "{scheme}://{netloc}".format(scheme=scheme, netloc=self.netloc)

    def set_base_dir(self, base_dir=None):

        if base_dir:
            self.base_dir = base_dir
        else:
            try:
                self.base_dir = join(os.environ['SAS_BASE_DIR'], '')
            except:
                pass

    def location(self, filetype, base_dir=None, **kwargs):
        """Return the location of the relative sas path of a given type of file.

        Parameters
        ----------
        filetype : str
            File type parameter.

        Returns
        -------
        full : str
            The relative sas path to the file.
        """

        full = kwargs.get('full', self.full(filetype, **kwargs))

        self.set_base_dir(base_dir=base_dir)
        location = full[len(self.base_dir):] if full and full.startswith(self.base_dir) else None

        if '//' in location:
            location = location.replace('//', '/')

        return location

    def url(self, filetype, base_dir=None, sasdir='sas', **kwargs):
        """Return the url of a given type of file.

        Parameters
        ----------
        filetype : str
            File type parameter.

        Returns
        -------
        full : str
            The sas url to the file.
        """

        location = self.location(filetype, **kwargs)
        return join(self.remote_base, sasdir, location) if self.remote_base and location else None


class Path(BasePath):
    """Derived class.  Sets a particular template file.
    """
    def __init__(self, mirror=False, public=False, verbose=False):
        try: tree_dir = os.environ['TREE_DIR']
        except KeyError:
            raise NameError("Could not find TREE_DIR in the environment!  Did you load the tree product?")
        pathfile = os.path.join(tree_dir, 'data', 'sdss_paths.ini')

        super(Path, self).__init__(pathfile, mirror=mirror, public=public, verbose=verbose)

    def plateid6(self, filetype, **kwargs):
        """Print plate ID, accounting for 5-6 digit plate IDs.

        Parameters
        ----------
        filetype : str
            File type parameter.
        plateid : int or str
            Plate ID number.  Will be converted to int internally.

        Returns
        -------
        plateid6 : str
            Plate ID formatted to a string of 6 characters.
        """
        plateid = int(kwargs['plateid'])
        if plateid < 10000:
            return "{:0>6d}".format(plateid)
        else:
            return "{:d}".format(plateid)

    def platedir(self, filetype, **kwargs):
        """Returns plate subdirectory in :envvar:`PLATELIST_DIR` of the form: ``NNNNXX/NNNNNN``.

        Parameters
        ----------
        filetype : str
            File type parameter.
        plateid : int or str
            Plate ID number.  Will be converted to int internally.

        Returns
        -------
        platedir : str
            Plate directory in the format ``NNNNXX/NNNNNN``.
        """
        plateid = int(kwargs['plateid'])
        plateid100 = plateid // 100
        subdir = "{:0>4d}".format(plateid100) + "XX"
        return os.path.join(subdir, "{:0>6d}".format(plateid))

    def spectrodir(self, filetype, **kwargs):
        """Returns :envvar:`SPECTRO_REDUX` or :envvar:`BOSS_SPECTRO_REDUX`
        depending on the value of `run2d`.

        Parameters
        ----------
        filetype : str
            File type parameter.
        run2d : int or str
            2D Reduction ID.

        Returns
        -------
        spectrodir : str
            Value of the appropriate environment variable.
        """
        if str(kwargs['run2d']) in ('26', '103', '104'):
            return os.environ['SPECTRO_REDUX']
        else:
            return os.environ['BOSS_SPECTRO_REDUX']

    def definitiondir(self, filetype, **kwargs):
        """Returns definition subdirectory in :envvar:`PLATELIST_DIR` of the form: ``NNNNXX``.

        Parameters
        ----------
        filetype : str
            File type parameter.
        designid : int or str
            Design ID number.  Will be converted to int internally.

        Returns
        -------
        definitiondir : str
            Definition directory in the format ``NNNNXX``.
        """

        designid = int(kwargs['designid'])
        designid100 = designid // 100
        subdir = "{:0>4d}".format(designid100) + "XX"
        return subdir


class AccessError(Exception):
    pass



import pytest
from wtforms.form import FormMeta

from wtforms_alchemy import (
    model_form_factory,
    model_form_meta_factory,
    ModelFormMeta
)


class _MetaWithInit(FormMeta):
    def __init__(cls, *args, **kwargs):
        cls.test_attr = 'SomeVal'
        FormMeta.__init__(cls, *args, **kwargs)


MetaWithInit = model_form_meta_factory(_MetaWithInit)


class _MetaWithoutInit(FormMeta):
    test_attr = 'SomeVal'


MetaWithoutInit = model_form_meta_factory(_MetaWithoutInit)


@pytest.fixture(params=[MetaWithInit, MetaWithoutInit, ModelFormMeta])
def model_form_all(request):
    """Returns one of each possible model form classes with custom and the
    original metaclass."""
    ModelForm = model_form_factory(meta=request.param)
    return ModelForm


@pytest.fixture(params=[MetaWithInit, MetaWithoutInit])
def model_form_custom(request):
    """Returns one of each possible model form classes with custom
    metaclasses."""
    return model_form_factory(meta=request.param)



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2016-10-11 13:24:56
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2016-10-11 23:55:08

from __future__ import print_function, division, absolute_import
import os
from collections import OrderedDict
from ConfigParser import SafeConfigParser


class Tree(object):
    ''' The Tree Class '''

    def __init__(self, *args, **kwargs):
        ''' initialize the sdss tree '''
        key = kwargs.get('key', None)
        self.setRoots()
        self.loadConfig()
        self.branchOut(limb=key)
        # add the general directories
        if key is not None:
            self.addPathsToOS(key='general')
        self.addPathsToOS(key=key)

    def __repr__(self):
        return ('Tree(sas_base_dir={0})'.format(self.sasbasedir))

    def setRoots(self):
        ''' Set the roots of the tree in the os environment '''

        # Check for TREE_DIR
        self.treedir = os.environ.get('TREE_DIR', None)
        if not self.treedir:
            treefilepath = os.path.dirname(os.path.abspath(__file__))
            self.treedir = treefilepath.rsplit('/', 2)[0]
            os.environ['TREE_DIR'] = self.treedir

        # Read the config file
        self.configfile = os.path.join(self.treedir, 'data', 'sdsswork.cfg')

        # Check sas_base_dir
        if 'SAS_BASE_DIR' in os.environ:
            self.sasbasedir = os.environ["SAS_BASE_DIR"]
        else:
            self.sasbasedir = os.path.expanduser('~/sas')

        # make the directories
        if not os.path.isdir(self.sasbasedir):
            os.makedirs(self.sasbasedir)

    def loadConfig(self):
        ''' load the sdsswork config file '''
        self._cfg = SafeConfigParser()
        self._cfg.read(self.configfile)
        # create the local tree environment
        self.environ = OrderedDict()
        self.environ['default'] = self._cfg.defaults()
        # set the filesystem envvar to sas_base_dir
        self._file_replace = '@FILESYSTEM@'
        if self.environ['default']['filesystem'] == self._file_replace:
            self.environ['default']['filesystem'] = self.sasbasedir

    def branchOut(self, limb=None):
        ''' Set the individual section branches

        This adds the various sections of the config file into the
        tree environment for access later. Optically can specify a specific
        section.

        Parameters:
            section (str/list):
                The name of the section of the config to add into the environ
                or a list of strings

        '''

        # Filter on sections
        if not limb:
            limbs = self._cfg.sections()
        else:
            # we must have the general always + secton
            limb = limb if type(limb) == list else [limb]
            limbs = ['general']
            limbs.extend(limb)

        # add all limbs into the tree environ
        for leaf in limbs:
            self.environ[leaf] = OrderedDict()
            options = self._cfg.options(leaf)
            for opt in options:
                if opt in self.environ['default']:
                    continue
                val = self._cfg.get(leaf, opt)
                if val.find(self._file_replace) == 0:
                    val = val.replace(self._file_replace, self.sasbasedir)
                self.environ[leaf][opt] = val

    def getPaths(self, key):
        ''' Retrieve a set of environment paths from the config

        Parameters:
            key (str):
                The section name to grab from the environment

        Returns:
            self.environ[newkey] (OrderedDict):
                An ordered dict containing all of the paths from the
                specified section, as key:val = name:path
        '''
        newkey = key if key in self.environ else key.upper() if key.upper() \
            in self.environ else None
        if newkey:
            return self.environ[newkey]
        else:
            raise KeyError('Key {0} not found in tree environment'.format(key))

    def addPathsToOS(self, key=None):
        ''' Add the paths in tree environ into the os environ

        This code goes through the tree environ and checks
        for existence in the os environ, then adds them

        Parameters:
            key (str):
                The section name to check against / add
        '''

        if key is not None:
            allpaths = key if type(key) == list else [key]
        else:
            allpaths = [k for k in self.environ.keys() if 'default' not in k]

        for key in allpaths:
            paths = self.getPaths(key)
            self.checkPaths(paths)

    def checkPaths(self, paths):
        ''' Check if the path is in the os environ, and if not add it

        Paramters:
            paths (OrderedDict):
                An ordered dict containing all of the paths from the
                a given section, as key:val = name:path
        '''
        for pathname, path in paths.items():
            if pathname.upper() not in os.environ:
                os.environ[pathname.upper()] = os.path.normpath(path)





#!/usr/bin/env python
# encoding: utf-8
"""

general.py

Licensed under a 3-clause BSD license.

Revision history:
    17 Feb 2016 J. Sánchez-Gallego
      Initial version
    18 Feb 2016 B. Cherinka
        Added buildRouteMap API call

"""

from __future__ import division
from __future__ import print_function
from flask_classy import route
from brain.api.base import BrainBaseView
from brain.core.exceptions import BrainError
from flask import url_for, current_app, jsonify
import json
try:
    from urllib import unquote
except ImportError:
    from urllib.parse import unquote


class BrainGeneralRequestsView(BrainBaseView):
    """A collection of requests for generic purposes."""

    route_base = '/general/'

    def index(self):
        res = {'data': 'this is a general Brain Function!'}
        self.update_results(res)
        return jsonify(self.results)

    @route('/getroutemap/', endpoint='getroutemap')
    def buildRouteMap(self):
        """ Build the URL route map for all routes in the Flask app.

        .. :quickref: General; Returns the urlmap dictionary of Marvin API routes

        Syntax of output:  {"api": {blueprint: {endpoint: {'methods':x, 'url':x} } }

        :form release: the release of MaNGA
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson json data: dictionary of returned data
        :json dict urlmap: dict of the Marvin API routes
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters
        :raises BrainError: Raised when url_for can't format the endpoint name into a valid url.

        **Example request**:

        .. sourcecode:: http

           GET /marvin2/api/general/getroutemap/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": {"urlmap": {"api": {"CubeView:index": {"methods": "HEAD,OPTIONS,GET","url": "/marvin2/api/cubes/"},
                                         ...
                                 }
                      }
           }

        """

        output = {}
        for rule in current_app.url_map.iter_rules():
            # get options
            options = {}
            for arg in rule.arguments:
                options[arg] = '[{0}]'.format(arg)
            options['_external'] = False
            # get endpoint
            fullendpoint = rule.endpoint
            esplit = fullendpoint.split('.')
            grp, endpoint = esplit[0], None if len(esplit) == 1 else esplit[1]
            output.setdefault(grp, {}).update({endpoint: {}})
            # get methods
            methods = ','.join(rule.methods)
            output[grp][endpoint]['methods'] = methods
            # build url
            try:
                rawurl = url_for(fullendpoint, **options)
            except ValueError as e:
                raise BrainError('Error generating url for {0}: {1}'.format(fullendpoint, e))
            url = unquote(rawurl).replace('[', '{').replace(']', '}')
            output[grp][endpoint]['url'] = url

        res = {'urlmap': output, 'status': 1}
        self.update_results(res)
        return jsonify(self.results)


# GeneralRequestsView.register(apiGeneral)
